#!/usr/bin/env python

import numpy as np
from spt3g import core
from spt3g.maps import FlatSkyMap, MapProjection, get_ra_dec_map, get_map_moments
from scipy.stats import skew, kurtosis

# Sparse extension operators
m = FlatSkyMap(500, 20, core.G3Units.arcmin)
m[7345] = 4
m[7345-500] = 4 # Backward
m[7345+500] = 4 # Forward
assert(m.sparse)
assert(m.npix_allocated == 3)
assert(m.npix_nonzero == 3)
m[7345-4*500] = 4 # Several steps back
assert(m.npix_allocated == 6)
assert(m.npix_nonzero == 4)
m[7345+3*500] = 4 # Several steps forward
assert(m.npix_allocated == 8)
assert(m.npix_nonzero == 5)

# Simple in-place operators
m = FlatSkyMap(500, 20, core.G3Units.arcmin)
m[15] = 10
assert(m.sparse)

m *= 5
m /= 25
assert(m[15] == 2)
assert(m.sparse)
assert(m[16] == 0)

assert((-m).sparse)
assert((-m)[15] == -2)

m += 3
m -= 14
assert(m[15] == -9)
assert(not m.sparse)
assert(m[16] == -11)

n = 1 - m
assert(n[15] == 10)
assert(n[16] == 12)

a = -11 * np.ones(m.shape)
a[0, 15] += 2
assert((m == a).all()) # Implicitly tests numpy conversions too

assert((m*0).npix_allocated == 0)

m += 11
m.sparse = True
assert(m.npix_allocated == 1)

n = 2. / m
assert(n[15] == 1)
assert(np.isinf(n[16]))
assert(n.npix_allocated == n.size)

# compactification
np.asarray(n)[np.isinf(n)] = np.nan
n.compact(zero_nans=True)
assert(n[16] == 0)
assert(n.npix_allocated == 1)

np.asarray(n)[np.isinf(n)] = np.nan
n.sparse = True
n.compact(zero_nans=True)
assert(n[16] == 0)
assert(n.npix_allocated == 1)

n = m ** 2.
assert(n[15] == 4)
assert(n.npix_allocated == 1)

# Map-by-map operations, with two sparse maps, one dense and one sparse,
# and two dense

m *= 2 # Get numbers bigger
assert((m == n).all())
assert((m > 0).any())
assert((m > 0).sum() == 1)
assert((m > 0).to_map().npix_allocated == 1)

m1 = m
m2 = m.copy()
m2.sparse = False

nm1 = m1.npix_allocated
nm2 = m2.npix_allocated

for pair in [(m1, m1), (m1, m2), (m2, m1), (m2, m2)]:
    t = pair[0] * pair[1]
    assert(t.sparse == pair[0].sparse)
    assert(t.npix_allocated == pair[0].npix_allocated)
    assert(t[15] == 16)
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)

    t = pair[0] + pair[1]
    assert(t.sparse == pair[0].sparse)
    assert(t.npix_allocated == pair[0].npix_allocated)
    assert(t[15] == 8)
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)

    t = pair[0] - 2 * pair[1]
    assert(t.sparse == pair[0].sparse)
    assert(t.npix_allocated == pair[0].npix_allocated)
    assert(t[15] == -4)
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)

    t = pair[0] / pair[1]
    assert(t.sparse == pair[0].sparse)
    assert(t.npix_allocated == t.size)
    assert(t[15] == 1)
    assert(not np.isfinite(t[12]))
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)

# With a null map
m3 = m.clone(False)

for pair in [(m1, m3), (m2, m3), (m3, m2), (m3, m1)]:
    nonnull = pair[1] if pair[0] is m3 else pair[0]

    t = pair[0] * pair[1]
    assert(t.sparse == True)
    assert(t.npix_allocated == 0)
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)
    assert(m3.npix_allocated == 0)

    t = pair[0] + pair[1]
    assert(t.sparse == nonnull.sparse)
    assert(t.npix_allocated == nonnull.npix_allocated)
    assert(t[15] == 4)
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)
    assert(m3.npix_allocated == 0)

    t = pair[0] - pair[1]
    assert(t.sparse == nonnull.sparse)
    assert(t.npix_allocated == nonnull.npix_allocated)
    assert(t[15] == -4 or t[15] == 4)
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)
    assert(m3.npix_allocated == 0)

    t = pair[0] / pair[1]
    assert(not np.isfinite(t[12]))
    if pair[0] is m3:
        assert(t[15] == 0)
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)
    assert(m3.npix_allocated == 0)


for shape in [(20, 500), (21, 501)]:
    print('shape', shape)

    # patch extraction / insertion
    m = FlatSkyMap(shape[1], shape[0], core.G3Units.arcmin, proj=MapProjection.ProjZEA)
    np.asarray(m)[:] = np.random.randn(*m.shape)
    malpha, mdelta = get_ra_dec_map(m)

    x0 = 45
    y0 = 13

    for dy, dx in [[10, 50], [11, 51]]:
        print('    patch', (dy, dx))
        p = m.extract_patch(45, 13, dx, dy)
        palpha, pdelta = get_ra_dec_map(p)
        x1 = x0 - dx // 2
        y1 = y0 - dy // 2
        sx = slice(x1, x1 + dx)
        sy = slice(y1, y1 + dy)
        assert(np.allclose(np.asarray(malpha)[sy, sx], palpha))
        assert(np.allclose(np.asarray(mdelta)[sy, sx], pdelta))
        assert(np.allclose(np.asarray(m)[sy, sx], p))

        m2 = m.clone(False)
        m2.insert_patch(p)
        assert(np.allclose(np.asarray(m)[sy, sx], np.asarray(m2)[sy, sx]))

    # Slice operators: make sure they work like numpy slicing
    assert((np.asarray(m[10:17,320:482]) == np.asarray(m.copy())[10:17,320:482]).all())
    # But give the right type...
    assert(m[10:17,320:482].__class__ == m.__class__)

    # Try setting things
    old_chunk = m[10:17,320:482]
    m[10:17,320:482] = old_chunk*2
    assert((np.asarray(m.copy())[10:17,320:482] == np.asarray(old_chunk)*2).all())
    m[10:17,320:482] = np.asarray(old_chunk*3)
    assert((np.asarray(m.copy())[10:17,320:482] == np.asarray(old_chunk)*3).all())
    m[10:17,320:482] = old_chunk
    mcopy = m.copy()
    m[:] = np.asarray(m * 2)
    assert((np.asarray(m) == np.asarray(mcopy * 2)).all())
    m[:] = np.asarray(mcopy)

    # Make sure inserting it in the wrong place (where coordinates don't make sense, but numpy
    # would allow it) fails
    failed = False
    try:
        m[11:18,320:482] = old_chunk
    except ValueError:
        failed = True
    assert(failed)

    # negative slice indices
    assert((np.asarray(m.copy())[-10:-3, -180:-18] == np.asarray(m.copy())[-10:-3, -180:-18]).all())

    # padding / cropping, with even and odd changes in dimension
    pad = 10
    for off in [0, 1]:
        print('    padding', 2 * pad + off)

        mpad = m.reshape(m.shape[1] + 2 * pad + off, m.shape[0] + 2 * pad + off)
        assert(mpad.npix_allocated == m.npix_allocated)
        a0 = np.array([m.alpha_center, m.delta_center])
        a1 = np.array([mpad.alpha_center, mpad.delta_center])
        assert(np.allclose(a0, a1))
        x0 = np.array([m.y_center, m.x_center])
        x1 = np.array([mpad.y_center, mpad.x_center])
        v0 = m[m.angle_to_pixel(*a0)]
        v1 = mpad[mpad.angle_to_pixel(*a1)]
        assert(v0 == v1)
        palpha, pdelta = get_ra_dec_map(mpad)
        sx = slice(mpad.shape[1] // 2 - m.shape[1] // 2, mpad.shape[1] // 2 - m.shape[1] // 2 + m.shape[1])
        sy = slice(mpad.shape[0] // 2 - m.shape[0] // 2, mpad.shape[0] // 2 - m.shape[0] // 2 + m.shape[0])
        assert(np.allclose(np.asarray(palpha)[sy, sx], malpha))
        assert(np.allclose(np.asarray(pdelta)[sy, sx], mdelta))
        assert(np.allclose(np.asarray(mpad)[sy, sx], np.asarray(m)))

        mcrop = mpad.reshape(m.shape[1], m.shape[0])
        calpha, cdelta = get_ra_dec_map(mcrop)
        a2 = np.array([mcrop.alpha_center, mcrop.delta_center])
        assert(np.allclose(a0, a2))
        x2 = np.array([mcrop.y_center, mcrop.x_center])
        assert(np.allclose(x0, x2))
        v2 = mcrop[mcrop.angle_to_pixel(*a2)]
        assert(v0 == v2)
        assert(np.allclose(calpha, malpha))
        assert(np.allclose(cdelta, mdelta))
        assert(np.allclose(mcrop, m))
        assert(m.compatible(mcrop))

        # statistics
        m1 = np.asarray(m).ravel()
        stats0 = [np.mean(m1), np.var(m1), skew(m1), kurtosis(m1)]
        stats1 = get_map_moments(m, order=4)
        assert(np.allclose(stats1, stats0))
        med0 = np.median(np.asarray(m1))
        med1 = np.median(m)
        assert(np.allclose(med1, med0))

        stats2 = get_map_moments(mpad, order=4, ignore_zeros=True)
        assert(np.allclose(stats2, stats0))
        med2 = mpad.median(where=mpad.to_mask())
        assert(np.allclose(med2, med0))

        mpad[mpad == 0] = np.nan
        stats3 = get_map_moments(mpad, order=4, ignore_nans=True)
        assert(np.allclose(stats3, stats0))
        med3 = mpad.median(where=~mpad.isnan())
        assert(np.allclose(med3, med0))

        mask = mpad.to_mask(zero_nans=True)
        stats4 = get_map_moments(mpad, mask=mask, order=4)
        assert(np.allclose(stats4, stats0))
        med4 = mpad.median(where=mask)
        assert(np.allclose(med4, med0))

# convolution
from scipy.signal import convolve2d
from spt3g.maps import convolve_map

kernel = FlatSkyMap(np.random.randn(5, 5), m.res)
mconv = convolve2d(m, kernel, mode='same')
mconv2 = convolve_map(m, kernel)
assert(np.allclose(mconv, mconv2))



#!/usr/bin/env python

import numpy
from spt3g import core

# Check that timestream maps being cast with numpy.asarray have the right
# content and are indistinguishable from numpy.asarray(list(tsm.values())

tsm = core.G3TimestreamMap()
start = core.G3Time.Now()
stop = start + 5*core.G3Units.s
keys = ['D', 'A', 'B', 'C']
for ts in keys:
	i = core.G3Timestream(numpy.random.normal(size=600))
	i.start = start
	i.stop = stop
	i.units = core.G3TimestreamUnits.Tcmb
	tsm[ts] = i

buffer1d = numpy.asarray(list(tsm.values()))

buffer2d = numpy.asarray(tsm)

numpy.testing.assert_array_equal(keys, list(tsm.keys()))
assert(buffer2d.shape == (4,600))
numpy.testing.assert_array_equal(buffer1d, buffer2d)

# Try building the TSM directly
tsm2 = core.G3TimestreamMap(tsm.keys(), buffer1d)
numpy.testing.assert_array_equal(numpy.asarray(tsm2), buffer1d)
numpy.testing.assert_array_equal(numpy.asarray(tsm2), buffer2d)

# Try building semi-directly, using a list of numpy arrays
tsm2 = core.G3TimestreamMap(tsm.keys(), [x for x in buffer1d])
numpy.testing.assert_array_equal(numpy.asarray(tsm2), buffer1d)
numpy.testing.assert_array_equal(numpy.asarray(tsm2), buffer2d)

# Round-trip to and from numpy
tsm2 = core.G3TimestreamMap(tsm.keys(), buffer2d)
numpy.testing.assert_array_equal(numpy.asarray(tsm2), buffer1d)
numpy.testing.assert_array_equal(numpy.asarray(tsm2), buffer2d)

# Now try writing to it
numpy.asarray(tsm2)[1,5] = 16
assert(numpy.asarray(tsm2)[1,5] == 16)

# And that writes propagate back to the underlying G3Timestream
buffer2d[1] = numpy.random.normal(size=600)
numpy.testing.assert_array_equal(numpy.asarray(tsm['B']), buffer2d[keys.index('B')])

# Test initialization from numpy and data copying/sharing
buffer1d[2,6] = -3.0

# By default (copy_data = True), changes should not propagate from the array used to initialize
# to the timestreams
tsm2 = core.G3TimestreamMap(tsm.keys(), buffer1d)
buffer1d[2,6] = 14.0
assert(numpy.asarray(tsm2)[2,6] != 14.0)

# But we should be able to set things up so that they do
tsm2 = core.G3TimestreamMap(tsm.keys(), buffer1d, copy_data = False)
buffer1d[2,6] = 16.0
assert(numpy.asarray(tsm2)[2,6] == 16.0)

# Test setting secondary properties at the time of direct construction
tsm2 = core.G3TimestreamMap(tsm.keys(), buffer1d, start=start, stop=stop,
                            units=core.G3TimestreamUnits.Counts,
                            compression_level=3, bit_depth=24)
for ts in tsm2.values():
	assert ts.start == start
	assert ts.stop == stop
	assert ts.units == core.G3TimestreamUnits.Counts
	assert ts.compression_level == 3
	assert ts.bit_depth == 24



from . import G3FrameType
from .modconstruct import g3decorator
from copy import copy
import inspect
import textwrap
import re

def get_function_signature(f, replacement_kwargs=None):
    '''
    Returns the function signature of f.  If replacement_kwargs is supplied
    replaces the default values with the ones in replacement_kwargs.
    '''
    sig = inspect.signature(f)

    if replacement_kwargs is not None:
        params = []
        for name, par in sig.parameters.items():
            if name in replacement_kwargs:
                par = par.replace(default=replacement_kwargs[name])
            params.append(par)
        sig = sig.replace(parameters=params)

    return f.__name__ + re.sub('<(.*) at (.*)>', '<\\1>', str(sig))


@g3decorator
class cache_frame_data(object):
    '''
    This is a decorator for use with G3Modules written as functions.  It enables
    a function to use cached values from other types of frames in the processing
    of a frame.
    
    To make that confusing sentence clearer with an example, in a lot of cases
    we want to have a module that works on Scan frames, but have access to the
    BolometerProperties.  This decorator allows you to specify the information
    to cache.  This case looks like:

    .. code-block:: python

        @core.cache_frame_data(type=core.G3FrameType.Scan, bolo_props='BolometerProperties')
        def FlagSomeStuff(frame, flag_key='Flags', bolo_props=None):
            pass

    You specify the type of frame the function is operating on with the type
    argument.  Any additional keyword arguments specifies information to cache
    and send to the function.

    For the keyword args passed to the decorator having the format: Key = Value.
    Key specifies the name of the argument that we pass the infromation to in
    the decorated function.  If the decorated function is called with Key as an
    argument it will overwrite the value specified in the decorator.
       
    Value specifies the default path to look for the cached data.  It will
    search all of the frames that do not have the frame type 'type' for a key
    with that value.  This can be overridden when calling the decorated
    function.
    '''

    def __init__(self_outer, type, **kwargs):
        self_outer.keyargs = kwargs
        self_outer.type = type
    def __call__(self_outer, f):
        func_sig = get_function_signature(f, replacement_kwargs = self_outer.keyargs)
        doc_prepend = '%s\n\n' % func_sig
        func_doc = textwrap.dedent(str(f.__doc__ or ''))
        doc_append = '''\n\n
This function automatic caches values from frames that contain the keys to be
cached.  It will only operate on %s frames.  If the cached keys are not found it
will pass None to the function.  You can change the default key by supplying the
following arguments:

Cached Values
-------------
''' % (str(self_outer.type or 'any type of') )
        for k,v in self_outer.keyargs.items():
            doc_append += '    %s = "%s"\n' % ( k, v )

        class WrappedFunc:
            def __init__(self, *args, **kwargs): 
                self.args = args
                self.kwargs = kwargs

                self.argument_map = copy(self_outer.keyargs)
                pop_ks = []
                for k in self.kwargs.keys():
                    if k in self.argument_map:
                        self.argument_map[k] = self.kwargs[k]
                        pop_ks.append(k)
                for k in pop_ks:
                    self.kwargs.pop(k)

            def __call__(self, frame):
                for vname, stored_key in self.argument_map.items():
                    if stored_key and stored_key in frame:
                        self.kwargs[vname] = frame[stored_key]
                if self_outer.type is None or frame.type == self_outer.type:
                    return f(frame, *(self.args), **(self.kwargs))
        WrappedFunc.__wrapped__ = f
        WrappedFunc.__name__ = f.__name__
        WrappedFunc.__doc__ = doc_prepend + func_doc + doc_append
        WrappedFunc.__g3module__ = True
        return WrappedFunc


@g3decorator
class scan_func_cache_data(cache_frame_data):
    '''
    This is a simple wrapper around cache_frame_data where the type argument has been set to
    core.G3FrameType.Scan.
    '''
    def __init__(self, **kwargs):
        super(scan_func_cache_data, self).__init__(G3FrameType.Scan, **kwargs)



import spt3g._libmaps
for k in dir(spt3g._libmaps):
    print(k, getattr(spt3g._libmaps, k))
from spt3g._libmaps import FlatSkyMap
import spt3g.maps
from spt3g import maps
from spt3g.maps import FlatSkyMap



#!/usr/bin/env python

import numpy as np
from spt3g import core
from spt3g.maps import HealpixSkyMap

# Simple in-place operators
m = HealpixSkyMap(64)
m[15] = 10
assert(m.ringsparse)

m *= 5
m /= 25
assert(m[15] == 2)
assert(m.ringsparse)
assert(m[16] == 0)

assert((-m).ringsparse)
assert((-m)[15] == -2)

m += 3
m -=14
assert(m[15] == -9)
assert(not m.ringsparse)
assert(m[16] == -11)

n = 1 - m
assert(n[15] == 10)
assert(n[16] == 12)

a = -11 * np.ones(m.shape)
a[15] += 2
assert((m == a).all()) # Implicitly tests numpy conversions too

assert((m * 0).npix_allocated == 0)

m += 11
m.indexedsparse = True
assert(m.npix_allocated == 1)
assert(m.npix_nonzero == 1)

n = 2. / m
assert(n[15] == 1)
assert(np.isinf(n[16]))
assert(n.npix_allocated == n.size)

# compactification
np.asarray(n)[np.isinf(n)] = np.nan
n.compact(zero_nans=True)
assert(n.ringsparse)
assert(n[16] == 0)
assert(n.npix_allocated == 1)
assert(n.npix_nonzero == 1)

np.asarray(n)[np.isinf(n)] = np.nan
n.ringsparse = True
n.compact(zero_nans=True)
assert(n.ringsparse)
assert(n[16] == 0)
assert(n.npix_allocated == 1)
assert(n.npix_nonzero == 1)

np.asarray(n)[np.isinf(n)] = np.nan
n.indexedsparse = True
n.compact(zero_nans=True)
assert(n.indexedsparse)
assert(n[16] == 0)
assert(n.npix_allocated == 1)
assert(n.npix_nonzero == 1)

m.ringsparse = True
n = m ** 2
assert(n[15] == 4)
assert(n.npix_allocated == 1)

# Map-by-map operations, with pairs of maps of any kind of density

m *= 2 # Get numbers bigger
assert((m == n).all())
assert((m > 0).any())
assert((m > 0).sum() == 1)
assert((m > 0).to_map().npix_allocated == 1)

m1 = m
m2 = m.copy()
m2.indexedsparse = True
m3 = m.copy()
m3.dense = True

nm1 = m1.npix_allocated
nm2 = m2.npix_allocated
nm3 = m3.npix_allocated

def check_state(a, b, check_size=True):
    assert(m1.npix_allocated == nm1)
    assert(m2.npix_allocated == nm2)
    assert(m3.npix_allocated == nm3)
    assert(a.indexedsparse == b.indexedsparse)
    assert(a.ringsparse == b.ringsparse)
    assert(a.dense == b.dense)
    if check_size:
        assert(a.npix_allocated == b.npix_allocated)

for pair in [(m1, m1), (m1, m2), (m1, m3),
             (m2, m1), (m2, m2), (m2, m3),
             (m3, m1), (m3, m2), (m3, m3)]:
    t = pair[0] * pair[1]
    check_state(pair[0], t)
    assert(t[15] == 16)

    t = pair[0] + pair[1]
    check_state(pair[0], t)
    assert(t[15] == 8)

    t = pair[0] - 2 * pair[1]
    check_state(pair[0], t)
    assert(t[15] == -4)

    t = pair[0] / pair[1]
    check_state(pair[0], t, False)
    assert(t.npix_allocated == t.size)
    assert(t[15] == 1)
    assert(not np.isfinite(t[12]))

# With a null map
m4 = m.clone(False)

for pair in [(m1, m4), (m2, m4), (m3, m4),
             (m4, m1), (m4, m2), (m4, m3)]:
    nonnull = pair[1] if pair[0] is m4 else pair[0]

    t = pair[0] * pair[1]
    assert(t.dense == False)
    assert(t.npix_allocated == 0)
    assert(m4.npix_allocated == 0)

    t = pair[0] + pair[1]
    check_state(t, nonnull)
    assert(t[15] == 4)
    assert(m4.npix_allocated == 0)

    t = pair[0] - pair[1]
    check_state(t, nonnull)
    assert(t[15] in [-4, 4])
    assert(m4.npix_allocated == 0)

    t = pair[0] / pair[1]
    assert(not np.isfinite(t[12]))
    if pair[0] is m4:
        assert(t[15] == 0)
    assert(m4.npix_allocated == 0)



from multiprocessing import get_context
import socket, pickle, errno, struct, time

from . import G3FrameType, G3Frame

# Require fork to avoid pickling errors
ctx = get_context("fork")

class Subproc(ctx.Process):
    '''
    Run a module in a subprocess, using python multiprocessing to proxy
    frames to it. If more than maxqueuelen frames are queued on the
    remote module, will block until there is space.
    '''
    def __init__(self, target=None, name=None, maxqueuelen=50):
        '''
        Set up a multiprocessing shim for the pipeline module <target>
        under the process name <name> (can be None).
        '''
        ctx.Process.__init__(self, name=name)
        self.targetmod = target

        self.queue = socket.socketpair()
        self.queue[0].setblocking(False)
        self.callsqueued = 0 # Count of frames outstanding at child

        self.maxqueued = maxqueuelen
        self.started = False
        self.outbuf = b''
    def __call__(self, frame):
        # Processing flow:
        # - Frames added to input queue (this will alert the remote
        #   process).
        # - Check output queue from the remote process and inject
        #   any frames there into the pipeline.

        if not self.started:
            self.started = True
            self.start()

        outbuf = pickle.dumps(frame)
        self.outbuf += struct.pack('i', len(outbuf)) + outbuf
        self.callsqueued += 1

        # Loop on output queue so long as
        # (a) we are expecting data back and
        # (b) (i) we need to finalize processing (an EndProcessing frame
        #     is here), (ii) we need to drain the queue, *or* (iii)
        #     there are pending frames from the remote end
        #
        # In general, attempt to make as much forward progress on both
        # sending and receiving data as possible (or necessary) in the
        # current moment. If we can't make progress, and don't need to,
        # save state and try again next time.

        allout = []
        inbuf = b''
        icount = 0
        while icount < 10 or \
         (frame.type == G3FrameType.EndProcessing and self.callsqueued > 0) or \
         self.callsqueued >= self.maxqueued:
            if len(self.outbuf) > 0:
                try:
                    resid = self.queue[0].send(self.outbuf)
                except socket.error as e:
                    # Just continue on EAGAIN
                    if e.errno != errno.EAGAIN:
                        raise
                else:
                    self.outbuf = self.outbuf[resid:]

            if self.callsqueued > 0:
                try:
                    inbuf = self.queue[0].recv(4)
                except socket.error as e:
                    if e.errno != errno.EAGAIN:
                        raise
                    # Fall through and do nothing if queue empty
                else:
                    # Unqueue a packet (length + data) from remote end
                    bytestoread = struct.unpack('i', inbuf)[0]
                    inbuf = b''
                    self.queue[0].setblocking(True)
                    while bytestoread > len(inbuf):
                        inbuf += self.queue[0].recv(bytestoread - len(inbuf))
                    self.queue[0].setblocking(False)
                                
                    allout += pickle.loads(inbuf)
                    self.callsqueued -= 1
                    icount = 0 # Forward progress! Try again quick

            icount += 1

        if frame.type == G3FrameType.EndProcessing:
            self.join()
        return allout
    def run(self):
        # Processing flow in worker process:
        # - Receive data from the remote end (blocking)
        # - Process data
        # - Return results, packaged as a list of frames
        #
        # Note that this must return *something* in response to every
        # frame to keep the outstanding frame accounting in the master
        # process working (an empty list is fine).

        while True:
            framelen = struct.unpack('i', self.queue[1].recv(4))[0]
            framebuf = b''
            while len(framebuf) < framelen:
                framebuf += self.queue[1].recv(framelen - len(framebuf))
            frame = pickle.loads(framebuf)

            outframes = self.targetmod(frame)

            if outframes is None:
                out = [frame]
            elif isinstance(outframes, G3Frame):
                out = [outframes]
            elif hasattr(outframes, '__iter__'):
                out = list(outframes)
            elif outframes:
                out = [frame]
            else:
                out = []

            outbuf = pickle.dumps(out)
            outbuf = struct.pack('i', len(outbuf)) + outbuf
            self.queue[1].send(outbuf)
                
            if frame.type == G3FrameType.EndProcessing:
                return




from .. import core
import random

@core.indexmod
class MPIFileReader(object):
    '''
    Do parallel I/O and processing across an MPI communicator. The style of
    parallelism here is that each process in the communicator gets a file, reads
    the file, and processes the file. Supports shared files, which are read
    from the lead process and then broadcasted to the others before they read
    their own.

    The list of files is specified as a list of lists. For example,
    ['offline_calibration.g3', ['0000.g3', '0002.g3', '0003.g3']]
    will cause all nodes to process 'offline_calibration.g3' and then the
    remaining files (in the parallel block denoted by the nested list)
    will be read and processed by the remaining nodes in a distributed way.
    '''
    def __init__(self, mpicomm, files):
        self.comm = mpicomm
        self.files = files
        self.cur_reader = None
        self.broadcast = False
    def nextreader(self):
        nextfilelist = self.files.pop(0)
        self.broadcast = False
        if isinstance(nextfilelist, str):
            self.broadcast = True
            if self.comm.rank == 0:
                self.cur_reader = core.G3Reader(nextfilelist)
            else:
                self.cur_reader = lambda fr: None
        else:
            toreadhere = nextfilelist[self.comm.rank::self.comm.size]
            self.cur_reader = core.G3Reader(toreadhere)
    def __call__(self, frame):
        assert(frame is None)
        out = []
        while len(out) == 0:
            if self.cur_reader is None:
                if len(self.files) == 0:
                    return []
                self.nextreader()
            out = self.cur_reader(frame)
            if self.broadcast:
                out = self.comm.bcast(out,root=0)
            if len(out) == 0:
                # Move to the next reader if this one is done
                self.cur_reader = None
        return out

@core.indexmod
class MPIFrameParallelizer(object):
    '''
    Do parallel I/O and processing across an MPI communicator. The style of
    parallelism here is that a set of IO processes read files from disk
    and distribute frames round-robin-style across a worker communicator.
    Frames will arrive in order on each process, with gaps between time
    segments, but no guarantees are made about relative order between
    processes. All CPU nodes receive all metadata frames.

    Rules to make this work:
    - First module on IO nodes must be MPIFileReader
    - Second module on IO nodes must be DeduplicateMetadata
    - Third module on IO nodes, first on workers, must be MPIFrameParallelizer

    You may want to use the MPIIODistributor pipe segment to make sure these
    rules are followed.
    '''
    def __init__(self, iocomm, cpucomm, cpucomm_startrank, dataframetype=[core.G3FrameType.Timepoint, core.G3FrameType.Scan]):
        '''
        Parellizes a frame stream across a communicator, distributing frames
        from M IO processes that are reading them to N CPU processes that
        are analyzing them. iocomm is a communicator containing exactly the
        set of IO nodes. cpucomm is a communicator containing (at least)
        all the IO and all the CPU nodes that the IO and CPU nodes can use
        to communicate with each other. The set of CPU nodes is the set of
        processes in cpucomm with rank greater than or equal to
        cpucomm_startrank. All frame types other than those in dataframetype
        appear on all CPU nodes, while frames of types in dataframetype
        are processed only by a single (random) CPU node.
        '''
        self.iocomm = iocomm
        self.cpucomm = cpucomm
        self.cpucomm_startrank = cpucomm_startrank
        self.dataframes = dataframetype
    def __call__(self, frame):
        if frame is None:
            # Initial frame is None on CPU nodes
            frame = self.cpucomm.recv()

            # This module (effectively) glues two G3Pipelines back-to-back.
            # G3Pipeline signals completion to modules with EndProcessing,
            # but completion is signaled *to* G3Pipeline with []. Translate
            # one to the other if needed.
            if frame.type == core.G3FrameType.EndProcessing:
                frame = []

            return frame
        if frame.type not in self.dataframes:
            # Synchronize and broadcast metadata; should be the same
            # in all streams
            if self.iocomm.rank == 0:
                checkframe = self.iocomm.bcast(frame, root=0)
            else:
                checkframe = self.iocomm.bcast(None, root=0)
            assert(checkframe.__getstate__()[0] == frame.__getstate__()[0]) # All metadata the same?
            # Get metadata frame in the queue on every worker process
            for i in list(range(self.cpucomm_startrank, self.cpucomm.size))[self.iocomm.rank::self.iocomm.size]:
                self.cpucomm.send(frame, i)
        else:
                # Send it somewhere random
                self.cpucomm.send(frame, random.randint(self.cpucomm_startrank, self.cpucomm.size-1))

        if frame.type == core.G3FrameType.EndProcessing:
            return frame

        return [] # Terminate processing on IO nodes
 
@core.pipesegment
def MPIIODistributor(pipe, mpicomm=None, n_io=10, files=[]):
    '''
    Read files from disk using the first n_io processes in mpicomm (COMM_WORLD
    by default), with processing of frames in those files occurring on the other
    processes in mpicomm. See documentation for MPIFileReader for the format of
    the files argument and MPIFrameParallelizer for information on the semantics
    of processing. Add this as the first module in your pipeline in place of
    core.G3Reader.
    '''
    if mpicomm is None:
        from mpi4py import MPI

        mpicomm = MPI.COMM_WORLD

    subcomm = mpicomm.Split(mpicomm.rank < n_io, mpicomm.rank)
    if mpicomm.rank < n_io:
        pipe.Add(MPIFileReader, mpicomm=subcomm, files=files)
        pipe.Add(core.util.DeduplicateMetadata)
    pipe.Add(MPIFrameParallelizer, iocomm=(subcomm if mpicomm.rank < n_io else None), cpucomm=mpicomm, cpucomm_startrank=n_io)




from .. import core
from . import HealpixSkyMap, FlatSkyMap, G3SkyMapWeights, G3SkyMapMask
from . import MapPolType, MapPolConv, MapCoordReference, MapProjection

import numpy as np
import os
import warnings

__all__ = [
    'load_skymap_fits',
    'save_skymap_fits',
    'load_skymapmask_fits',
    'save_skymapmask_fits',
    'SaveMapFrame',
]


@core.usefulfunc
def load_skymap_fits(filename, hdu=None, keys=None, memmap=False, apply_units=False):
    """
    Load a fits file containing a sky map.

    Arguments
    ---------
    filename : str
        Path to fits file
    hdu : int, optional
        If supplied, the data are extract from the given HDU index.
    keys : list of strings, optional
        If supplied, return only these keys in the output dictionary.
        Options are: T, Q, U, W.
    memmap : bool, optional
        Argument passed to astropy.io.fits.open. If True, the map is not read
        into memory, but only the required pixels are read when needed. Default:
        False.
    apply_units : bool, optional
        If True, and input maps have known units, multiply by the appropriate
        conversion factor to return maps in G3Units.

    Returns
    -------
    a dictionary of maps keyed with e.g. 'T', 'Q', 'U' and 'W'.
    """

    import astropy.io.fits
    assert(os.path.exists(filename))

    map_type = None
    pol = None
    map_opts = {}
    output = {}

    # defaults for missing header entries
    pol_conv = None
    units = 'Tcmb'
    coord_ref = 'Equatorial'
    weighted = False

    unit_dict = {
        'k_cmb': ('Tcmb', core.G3Units.K),
        'kcmb': ('Tcmb', core.G3Units.K),
        'kcmb^2': ('Tcmb', core.G3Units.K ** 2),
    }

    if keys is None:
        keys = ['T', 'Q', 'U', 'W']

    with astropy.io.fits.open(filename, memmap=memmap) as hdulist:
        for hidx, H in enumerate(hdulist):

            hdr = H.header
            mtype = hdr.get('MAPTYPE', None)

            if mtype == 'FLAT' or 'PROJ' in hdr or 'WCSAXES' in hdr or 'CTYPE1' in hdr:
                # flat map
                if not map_type:
                    map_type = 'flat'
                elif map_type != 'flat':
                    raise ValueError(
                        "Expected a {} sky map in HDU {}, found a flat map".format(
                            map_type, hidx
                        )
                    )
                # expect that flat sky maps on disk are probably in IAU
                if not pol_conv:
                    pol_conv = 'IAU'
            elif mtype == 'HEALPIX' or hdr.get('PIXTYPE', None) == 'HEALPIX' or 'NSIDE' in hdr:
                # healpix map
                if not map_type:
                    map_type = 'healpix'
                elif map_type != 'healpix':
                    raise ValueError(
                        "Expected a {} sky map in HDU {}, found a healpix map".format(
                            map_type, hidx
                        )
                    )
                # expect that healpix maps on disk are probably in COSMO
                if not pol_conv:
                    pol_conv = 'COSMO'

            if 'POLAR' in hdr:
                pdict = {'T': True, 'F': False}
                pol = pdict.get(hdr['POLAR'], hdr['POLAR'])
            if pol and map_type is not None:
                if hdr.get('POLCCONV', '').upper() not in ['IAU', 'COSMO']:
                    core.log_warn('Polarization convention not set, assuming %s' % pol_conv,
                                  unit='load_skymap_fits')
                else:
                    pol_conv = hdr['POLCCONV'].upper()
            uconv = None
            if 'TUNIT' in hdr:
                u = hdr['TUNIT'].lower()
                if u in unit_dict:
                    units, uconv = unit_dict[u]
            else:
                units = hdr.get('UNITS', units)
            if 'COORDSYS' in hdr:
                cdict = {'C': 'Equatorial', 'G': 'Galactic', 'L': 'Local'}
                coord_ref = cdict.get(hdr['COORDSYS'], coord_ref)
            else:
                coord_ref = hdr.get('COORDREF', coord_ref)
            overflow = hdr.get('OVERFLOW', 0)

            map_opts.update(
                units=getattr(core.G3TimestreamUnits, units),
                coord_ref=getattr(MapCoordReference, coord_ref),
                pol_conv=getattr(MapPolConv, pol_conv) if pol_conv else None,
            )

            if map_type == 'flat':
                map_opts.update(flat_pol=hdr.get('FLATPOL', False))
                map_opts.update(parse_wcs_header(hdr))

            elif map_type == 'healpix':
                nside = hdr['NSIDE']
                nested = hdr['ORDERING'].strip().lower() in ['nest', 'nested']
                map_opts.update(nested=nested)

            # primary HDU
            if H.data is None:
                continue

            # extracting a particular HDU
            if map_type is not None and hdu is not None and hidx != hdu:
                continue

            if map_type == 'flat':
                if hdr.get('ISWEIGHT', None):
                    if 'W' not in keys:
                        continue
                else:
                    if hdr.get('POLTYPE', 'T') not in keys:
                        continue
                data = np.array(H.data, dtype=float)
                if map_opts.pop('transpose', False):
                    data = np.array(data.T)
                if uconv is not None:
                    data *= uconv

            # map type must be known if the HDU contains map data
            if map_type not in ['flat', 'healpix']:
                raise ValueError("Unknown map type in HDU {}".format(hidx))

            if map_type == 'flat' and hdr.get('ISWEIGHT', None):
                # flat map weights
                assert('T' in output)
                if 'W' not in output:
                    output['W'] = G3SkyMapWeights()
                weight_map = output['W']
                pol_type = getattr(MapPolType, hdr['WTYPE'], None)
                fm = FlatSkyMap(data, pol_type=pol_type, **map_opts)
                fm.overflow = overflow
                setattr(weight_map, hdr['WTYPE'], fm)
                del data

            elif map_type == 'flat' and not hdr.get('ISWEIGHT', None):
                # flat map data
                ptype = hdr.get('POLTYPE', 'T')
                pol_type = getattr(MapPolType, ptype, None)
                fm = FlatSkyMap(data, pol_type=pol_type, **map_opts)
                fm.overflow = overflow
                output[ptype] = fm
                del data

            elif map_type == 'healpix':
                # healpix map data

                col_dict = {
                    'TEMPERATURE': 'T',
                    'Q_POLARISATION': 'Q',
                    'U_POLARISATION': 'U',
                    'Q-POLARISATION': 'Q',
                    'U-POLARISATION': 'U',
                    'I_STOKES': 'T',
                    'Q_STOKES': 'Q',
                    'U_STOKES': 'U',
                    'I': 'T',
                    'II': 'TT',
                    'IQ': 'TQ',
                    'IU': 'TU',
                    'II_COV': 'TT',
                    'IQ_COV': 'TQ',
                    'IU_COV': 'TU',
                    'QQ_COV': 'QQ',
                    'QU_COV': 'QU',
                    'UU_COV': 'UU',
                    'I_MEAN': 'T',
                    'Q_MEAN': 'Q',
                    'U_MEAN': 'U',
                }

                pix = None

                partial = hdr.get('INDXSCHM') == 'EXPLICIT' or hdr.get('OBJECT') == 'PARTIAL'

                for cidx, hcol in enumerate(H.data.names):
                    col = col_dict.get(hcol, hcol)
                    if col in 'TQU' and col not in keys:
                        continue
                    elif col in ['TT', 'TQ', 'TU', 'QQ', 'QU', 'UU'] and 'W' not in keys:
                        continue

                    data = np.array(H.data[hcol], dtype=float).ravel()

                    if col == 'PIXEL' or (partial and cidx == 0):
                        pix = np.array(data, dtype=int).ravel()
                        del data
                        continue

                    uconv = None
                    u = 'TUNIT{:d}'.format(cidx + 1)
                    if u in hdr:
                        u = hdr[u].lower()
                        if u in unit_dict:
                            units, uconv = unit_dict[u.lower()]
                            if uconv is not None:
                                data *= uconv
                    overflow = hdr.get('TOFLW{:d}'.format(cidx + 1), overflow)

                    weighted = hdr.get(
                        'TISWGT{:d}'.format(cidx + 1),
                        col in ['TT', 'TQ', 'TU', 'QQ', 'QU', 'UU'],
                    )
                    pol_type = getattr(MapPolType, col, None)
                    mdata = (pix, data, nside) if pix is not None else (data,)

                    if weighted:
                        assert('T' in output)
                        if 'W' not in output:
                            output['W'] = G3SkyMapWeights()
                        weight_map = output['W']
                        hm = HealpixSkyMap(*mdata, pol_type=pol_type, **map_opts)
                        hm.overflow = overflow
                        setattr(weight_map, col, hm)
                    else:
                        hm = HealpixSkyMap(*mdata, pol_type=pol_type, **map_opts)
                        output[col] = hm

                    del mdata, data

                del pix
                break

            del H.data

    for k, m in output.items():
        m.pol_conv = map_opts['pol_conv']
        if k == 'W':
            continue
        m.weighted = 'W' in output

    return output


@core.usefulfunc
def load_skymapmask_fits(filename, hdu=None, memmap=False):
    """
    Load a fits file containing a binary sky map mask.  Floating pointing or integer
    data are converted to booleans.  Assumes the fits file contains a single mask,
    with flat sky data stored in a single ``ImageHDU``, and healpix masks stored in a
    single ``BinTableHDU`` in full-sky format with implicit indexing over all pixels.

    Arguments
    ---------
    filename : str
        Path to fits file
    hdu : int, optional
        If supplied, the data are extract from the given HDU index.
    memmap : bool, optional
        Argument passed to astropy.io.fits.open. If True, the map is not read into
        memory, but only the required pixels are read when needed. Default: False.

    Returns
    -------
    G3SkyMapMask :
        The binary mask.
    """
    import astropy.io.fits

    map_type = None
    coord_ref = "Equatorial"
    map_opts = {}

    with astropy.io.fits.open(filename, memmap=memmap) as hdulist:
        for hidx, H in enumerate(hdulist):

            hdr = H.header
            mtype = hdr.get('MAPTYPE', None)

            if mtype == 'FLAT' or 'PROJ' in hdr or 'WCSAXES' in hdr or 'CTYPE1' in hdr:
                map_type = 'flat'
            elif mtype == 'HEALPIX' or hdr.get('PIXTYPE', None) == 'HEALPIX' or 'NSIDE' in hdr:
                map_type = 'healpix'

            if 'COORDSYS' in hdr:
                cdict = {'C': 'Equatorial', 'G': 'Galactic', 'L': 'Local'}
                coord_ref = cdict.get(hdr['COORDSYS'], coord_ref)
            else:
                coord_ref = hdr.get('COORDREF', coord_ref)
            map_opts.update(coord_ref=getattr(MapCoordReference, coord_ref))

            if map_type == 'flat':
                map_opts.update(parse_wcs_header(hdr))
            elif map_type == 'healpix':
                nside = hdr['NSIDE']
                nested = hdr['ORDERING'].strip().lower() in ['nest', 'nested']
                map_opts.update(nested=nested)

            # primary HDU
            if H.data is None:
                continue

            # extracting a particular HDU
            if map_type is not None and hdu is not None and hidx != hdu:
                continue

            # map type must be known if the HDU contains map data
            if map_type not in ['flat', 'healpix']:
                raise ValueError("Unknown map type in HDU {}".format(hidx))

            data = np.array(H.data, dtype=bool)

            if map_type == 'flat':
                # flat map data
                if map_opts.pop('transpose', False):
                    data = np.array(data.T)
                parent = FlatSkyMap(*data.shape, **map_opts)

            elif map_type == 'healpix':
                # healpix map data
                parent = HealpixSkyMap(nside, **map_opts)

            mask = G3SkyMapMask(parent, data.ravel())

            del H.data, data
            return mask


def load_proj_dict(inverse=False):
    """
    Dictionary that associates projection names to their three-letter codes in WCS.
    """

    projdict = {
        'Proj8': '!!!',
        'ProjSFL': 'SFL',
        'ProjCAR': 'CAR',
        'ProjSIN': 'SIN',
        'ProjARC': 'ARC',
        'ProjSTG': 'STG',
        'ProjZEA': 'ZEA',
        'ProjTAN': 'TAN',
        'ProjCEA': 'CEA',
        'ProjBICEP': 'CAR',
        'ProjHealpix': 'HPX',
    }

    if inverse:
        projdict_inverse = {}
        for k, v in projdict.items():
            projdict_inverse.setdefault(v, []).append(k)
        return projdict_inverse

    return projdict


def get_wcs(skymap, reset=False):
    """
    Creates WCS (world coordinate system) information from information in the
    input FlatSkyMap object.
    """

    if hasattr(skymap, '_wcs') and not reset:
        return skymap._wcs

    projdict = load_proj_dict()
    proj_abbr = projdict[str(skymap.proj)]

    from astropy.wcs import WCS

    w = WCS(naxis=2)

    if skymap.coord_ref == MapCoordReference.Equatorial:
        w.wcs.ctype = ['RA---{}'.format(proj_abbr), 'DEC--{}'.format(proj_abbr)]
        w.wcs.radesys = 'FK5'
        w.wcs.equinox = 2000
    elif skymap.coord_ref == MapCoordReference.Galactic:
        w.wcs.ctype = ['GLON-{}'.format(proj_abbr), 'GLAT-{}'.format(proj_abbr)]
    elif skymap.coord_ref == MapCoordReference.Local:
        w.wcs.ctype = ['RA---{}'.format(proj_abbr), 'DEC--{}'.format(proj_abbr)]
        w.wcs.radesys = 'ALTAZ'

    w.wcs.cdelt = [
        -skymap.x_res / core.G3Units.deg,
        skymap.res / core.G3Units.deg,
    ]
    w.wcs.cunit = ['deg', 'deg']
    w.wcs.crpix = [skymap.x_center + 1, skymap.y_center + 1]
    w.wcs.crval = [
        skymap.alpha_center / core.G3Units.deg,
        skymap.delta_center / core.G3Units.deg,
    ]

    if proj_abbr in ['SFL', 'CAR', 'CEA']:
        w.wcs.crval[1] = 0.0
        if proj_abbr == 'CEA':
            v = np.sin(skymap.delta_center / core.G3Units.rad) / skymap.y_res
        else:
            v = skymap.delta_center / skymap.y_res
        w.wcs.crpix[1] -= v
        if str(skymap.proj) == 'ProjBICEP':
            w.wcs.pc[0][0] = 1. / np.cos(skymap.delta_center / core.G3Units.rad)

    skymap._wcs = w
    return w

# set property
wcs_doc = "astropy.wcs.WCS instance containing projection information"
setattr(FlatSkyMap, 'wcs', property(get_wcs, doc=wcs_doc))


def create_wcs_header(skymap):
    """
    Return a FITS header for the WCS information in the FlatSkyMap object.
    """
    header = skymap.wcs.to_header()
    header['PROJ'] = str(skymap.proj)
    header['ALPHA0'] = skymap.alpha_center / core.G3Units.deg
    header['DELTA0'] = skymap.delta_center / core.G3Units.deg
    header['X0'] = skymap.x_center
    header['Y0'] = skymap.y_center
    return header


def parse_wcs_header(header):
    """
    Extract flat sky map keywords from a WCS fits header.  Raises an error if
    the header is malformed.
    """

    from astropy.wcs import WCS
    w = WCS(header)

    # parse projection
    ctype = w.wcs.ctype
    wcsproj = ctype[0][-3:]
    proj = header.get('PROJ', None)
    projdict = load_proj_dict(inverse=True)
    if wcsproj not in projdict or wcsproj == '!!!':
        raise ValueError('Unknown WCS projection {}'.format(wcsproj))
    projopts = projdict[wcsproj]
    if proj is None:
        proj = projopts[0]
    elif proj not in projopts:
        raise ValueError('PROJ keyword inconsistent with WCS coordinate type')

    # parse coordinate system
    coord_ref = None
    if ctype[0].startswith('RA-') or ctype[0].startswith('DEC-'):
        if w.wcs.radesys == 'FK5':
            coord_ref = MapCoordReference.Equatorial
        elif w.wcs.radesys == 'ALTAZ':
            coord_ref = MapCoordReference.Local
    elif ctype[0].startswith('GLON-') or ctype[0].startswith('GLAT-'):
        coord_ref = MapCoordReference.Galactic

    # parse coordinate order
    if ctype[0].startswith('RA-') or ctype[0].startswith('GLON-'):
        order = [0, 1]
        transpose = False
    elif ctype[0].startswith('DEC-') or ctype[0].startswith('GLAT-'):
        order = [1, 0]
        transpose = True
    else:
        raise ValueError('Unable to determine WCS coordinate axes')

    # parse resolution
    cdelt = w.wcs.get_cdelt()
    x_res = -cdelt[order[0]] * core.G3Units.deg
    res = cdelt[order[1]] * core.G3Units.deg

    # parse map center
    crval = w.wcs.crval
    alpha_center = header.get('ALPHA0', crval[order[0]]) * core.G3Units.deg
    delta_center = header.get('DELTA0', crval[order[1]]) * core.G3Units.deg

    crpix = w.wcs.crpix
    x_center = header.get('X0', crpix[order[0]] - 1)
    y_center = header.get('Y0', crpix[order[1]] - 1)

    if proj == 'ProjBICEP':
        v = 1. / np.cos(delta_center / core.G3Units.rad)
        if not np.isclose(v, w.wcs.pc[order[0]][order[0]]):
            raise ValueError("Inconsistent BICEP projection resolution")

    if wcsproj  == 'CEA':
        for i, m, v in w.wcs.get_pv():
            if i == order[1] + 1 and m == 1:
                if not np.isclose(v, 1):
                    raise ValueError(
                        "CEA projection with non-conformal scaling parameter "
                        "PV%d_1 not supported" % i
                    )

    # construct arguments
    map_opts = dict(
        proj=getattr(MapProjection, proj),
        res=res,
        x_res=x_res,
        alpha_center=alpha_center,
        delta_center=delta_center,
        x_center=x_center,
        y_center=y_center,
        transpose=transpose,
    )
    if coord_ref is not None:
        map_opts.update(coord_ref=coord_ref)

    return map_opts


@core.usefulfunc
def save_skymap_fits(filename, T, Q=None, U=None, W=None, overwrite=False,
                     compress=False, quantize_level=16.0, hdr=None):
    """
    Save G3 map objects to a fits file.

    `FlatSkyMap` objects are stored in a series of (optionally compressed)
    `ImageHDU` entries, in which each HDU contains the projection information in
    its header in standard WCS format, along with the image data for a single
    map (one of the Stokes maps or a weight map component).

    `HealpixSkyMap` objects are stored in a `BinTableHDU` extension, which
    contains the necessary header information for compatiblity with healpix map
    readers (e.g. `healpix.read_map`), and a single table with one column per
    Stokes map or weight map component.  Sparse maps are stored as cut-sky
    pixel-indexed tables, while dense maps are stored with implicit indexing
    over all pixels.  The former produces output that is equivalent to using
    `healpy.write_map` with the `partial=True` option.

    Arguments
    ---------
    filename : str
        Path to output file.  Must not exist, unless overwrite is True.
    T, Q, U : FlatSkyMap or HealpixSkyMap
        Maps to save
    W : G3SkyMapWeights
        Weights to save with the maps
    overwrite : bool
        If True, any existing file with the same name will be ovewritten.
    compress : str or bool
        If defined, and if input maps are FlatSkyMap objects, store these in a
        series of compressed image HDUs, one per map.  Otherwise, store input
        maps in a series of standard ImageHDUs, which are readable with older
        FITS readers (e.g. idlastro). If defined, the compression algorithm to
        be used by the Astropy class astropy.io.fits.CompImageHDU.
        Can be: 'RICE_1', 'RICE_ONE', 'PLIO_1', 'GZIP_1', 'GZIP_2' or
        'HCOMPRESS_1'. Only GZIP_1 and GZIP_2 are lossless, although only
        for integer data.
    quantize_level : float
        Floating point quantization level for compression.  Higher values result
        in more accurate floating point representation, but worse compression
        ratio.  See the astropy FITS image documention for details:
        https://docs.astropy.org/en/stable/io/fits/api/images.html
    hdr  : dict
       If defined, extra keywords to be appened to the FITS header. The dict
       can contain entries such as ``hdr['NEWKEY'] = 'New value'`` or
       ``hdr['NEWKEY'] = ('New value', "Comment for New value")``.
    """

    import astropy.io.fits

    if isinstance(T, (FlatSkyMap, HealpixSkyMap)):
        flat = isinstance(T, FlatSkyMap)
    else:
        raise TypeError("Input map must be a FlatSkyMap or HealpixSkyMap instance")

    ctype = None
    if compress == True:
        ctype = 'GZIP_2'
    elif isinstance(compress, str):
        ctype = compress

    if Q is not None:
        assert(U is not None)
        assert(U.pol_conv == MapPolConv.IAU or U.pol_conv == MapPolConv.COSMO)
        assert(T.compatible(Q))
        assert(T.compatible(U))
        pol = True
        maps = [T, Q, U]
        if flat:
            names = ['T', 'Q', 'U']
        else:
            names = ['TEMPERATURE', 'Q_POLARISATION', 'U_POLARISATION']
    else:
        assert(U is None)
        pol = False
        maps = [T]
        names = ['T' if flat else 'TEMPERATURE']

    if flat:
        header = create_wcs_header(T)

        bitpix = {
            np.dtype(np.int16): 16,
            np.dtype(np.int32): 32,
            np.dtype(np.float32): -32,
            np.dtype(np.float64): -64,
        }
        header['BITPIX'] = bitpix[np.asarray(T).dtype]
        header['NAXIS'] = 2

    else:
        header = astropy.io.fits.Header()

        header['PIXTYPE'] = 'HEALPIX'
        header['ORDERING'] = 'NEST' if T.nested else 'RING'
        header['NSIDE'] = T.nside
        if T.dense:
            header['INDXSCHM'] = 'IMPLICIT'
            header['OBJECT'] = 'FULLSKY'
        else:
            header['INDXSCHM'] = 'EXPLICIT'
            header['OBJECT'] = 'PARTIAL'

        cdict = {'Equatorial': 'C', 'Galactic': 'G', 'Local': 'L'}
        header['COORDSYS'] = cdict[str(T.coord_ref)]
        if header['COORDSYS'] == 'C':
            header['RADESYS'] = 'FK5'
            header['EQUINOX'] = 2000.0

        conv = {
            np.dtype(np.int16): 'I',
            np.dtype(np.int32): 'J',
            np.dtype(np.int64): 'K',
            np.dtype(np.float32): 'E',
            np.dtype(np.float64): 'D',
        }
        pix = None

    header['MAPTYPE'] = 'FLAT' if flat else 'HEALPIX'
    header['COORDREF'] = str(T.coord_ref)
    if pol:
        header['POLCCONV'] = str(U.pol_conv)
    header['POLAR'] = pol
    header['UNITS'] = str(T.units)
    header['WEIGHTED'] = T.weighted

    # Append extra metadata in hdr if defined
    if hdr:
        for keyword in hdr.keys():
            header[keyword] = hdr[keyword]

    hdulist = astropy.io.fits.HDUList()
    cols = astropy.io.fits.ColDefs([])

    try:
        for m, name in zip(maps, names):
            if flat:
                if compress:
                    hdu = astropy.io.fits.CompImageHDU(
                        np.asarray(m),
                        header=header,
                        compression_type=ctype,
                        quantize_level=quantize_level,
                    )
                else:
                    hdu = astropy.io.fits.ImageHDU(np.asarray(m), header=header)
                hdu.header['ISWEIGHT'] = False
                hdu.header['POLTYPE'] = name
                hdu.header['OVERFLOW'] = m.overflow
                hdu.header['FLATPOL'] = m.flat_pol
                hdulist.append(hdu)
                del hdu
            else:
                if not T.dense:
                    pix1, data = m.nonzero_pixels()
                    pix1 = np.asarray(pix1)
                    data = np.asarray(data)
                    data = data[np.argsort(pix1)]
                    pix1.sort()

                    if pix is None:
                        pix = pix1

                        fmt = conv.get(np.min_scalar_type(-pix.max()), 'K')
                        col = astropy.io.fits.Column(
                            name='PIXEL', format=fmt, array=pix, unit=None
                        )
                        cols.add_col(col)
                        del col
                    else:
                        # XXX different sparse maps can have different nonzero pixels
                        assert(len(pix) == len(pix1) and (pix == pix1).all())
                    del pix1

                else:
                    data = np.asarray(m)

                fmt = conv.get(data.dtype, 'D')
                col = astropy.io.fits.Column(
                    name=name, format=fmt, array=data, unit=str(m.units)
                )
                cols.add_col(col)
                del col
                del data

                idx = len(cols)
                header['TISWGT{:d}'.format(idx)] = False
                header['TOFLW{:d}'.format(idx)] = m.overflow

        if W is not None:
            if pol:
                assert(W.polarized)
                wnames = ['TT', 'TQ', 'TU', 'QQ', 'QU', 'UU']
            else:
                assert(not W.polarized)
                wnames = ['TT']

            for wt in wnames:
                m = getattr(W, wt)
                assert T.compatible(m), 'Weight {} is not compatible with T'.format(wt)

                if flat:
                    if compress:
                        hdu = astropy.io.fits.CompImageHDU(
                            np.asarray(m),
                            header=header,
                            compression_type=ctype,
                            quantize_level=quantize_level,
                        )
                    else:
                        hdu = astropy.io.fits.ImageHDU(np.asarray(m), header=header)
                    hdu.header['ISWEIGHT'] = True
                    hdu.header['WTYPE'] = wt
                    hdu.header['OVERFLOW'] = m.overflow
                    hdulist.append(hdu)
                    del hdu
                else:
                    if not T.dense:
                        pix1, data = m.nonzero_pixels()
                        pix1 = np.asarray(pix1)
                        data = np.asarray(data)
                        data = data[np.argsort(pix1)]
                        pix1.sort()
                        assert(len(pix) == len(pix1) and (pix == pix1).all())
                        del pix1
                    else:
                        data = np.asarray(m)

                    fmt = conv.get(data.dtype, 'D')
                    col = astropy.io.fits.Column(
                        name=wt, format=fmt, array=data, unit=str(m.units)
                    )
                    cols.add_col(col)
                    del col
                    del data

                    idx = len(cols)
                    header['TISWGT{:d}'.format(idx)] = False
                    header['TOFLW{:d}'.format(idx)] = m.overflow

        if not flat:
            del pix
            hdu = astropy.io.fits.BinTableHDU.from_columns(cols, header=header)
            hdulist.append(hdu)
            del hdu

        if overwrite:
            if os.path.exists(filename):
                os.remove(filename)
        else:
            assert(not os.path.exists(filename))

        hdulist.writeto(filename)

    finally:
        for col in cols.names:
            del cols[col].array
            cols.del_col(col)
        del cols
        for hdu in hdulist:
            del hdu.data
        del hdulist


@core.usefulfunc
def save_skymapmask_fits(filename, mask, overwrite=False, compress=False, hdr=None):
    """
    Save G3 mask objects to a fits file.

    Masks with ``FlatSkyMap`` parents are stored in a (optionally compressed)
    ``ImageHDU`` entry, which contains the projection information in its header in
    standard WCS format, along with the image data for a single mask.

    Masks with ``HealpixSkyMap`` parents are stored in a ``BinTableHDU`` extension,
    which contains the necessary header information for compatiblity with healpix map
    readers (e.g. `healpix.read_map`), and a single table with one column for the
    mask.  Masks are stored densely with implicit indexing over all pixels.

    Arguments
    ---------
    filename : str
        Path to output file.  Must not exist, unless overwrite is True.
    mask : G3SkyMapMask
        Mask to save
    overwrite : bool
        If True, any existing file with the same name will be ovewritten.
    compress : str or bool
        If defined, and if input mask has a ``FlatSkyMap`` parent, store this in a
        compressed image HDU.  Otherwise, store input mask in a standard ImageHDU,
        which is readable with older FITS readers (e.g. idlastro). If defined, the
        compression algorithm to be used by the Astropy class
        astropy.io.fits.CompImageHDU.  Can be: 'RICE_1', 'RICE_ONE', 'PLIO_1',
        'GZIP_1', 'GZIP_2' or 'HCOMPRESS_1'. Only GZIP_1 and GZIP_2 are lossless,
        although only for integer data.
    hdr  : dict
       If defined, extra keywords to be appened to the FITS header. The dict
       can contain entries such as ``hdr['NEWKEY'] = 'New value'`` or
       ``hdr['NEWKEY'] = ('New value', "Comment for New value")``.
    """
    import astropy.io.fits

    flat = isinstance(mask.parent, FlatSkyMap)

    if flat:
        header = create_wcs_header(mask.parent)
        header["BITPIX"] = 8
        header['NAXIS'] = 2

    else:
        header = astropy.io.fits.Header()

        header['PIXTYPE'] = 'HEALPIX'
        header['ORDERING'] = 'NEST' if mask.parent.nested else 'RING'
        header['NSIDE'] = mask.parent.nside
        header['INDXSCHM'] = 'IMPLICIT'
        header['OBJECT'] = 'FULLSKY'
        cdict = {'Equatorial': 'C', 'Galactic': 'G', 'Local': 'L'}
        header['COORDSYS'] = cdict[str(mask.parent.coord_ref)]
        if header['COORDSYS'] == 'C':
            header['RADESYS'] = 'FK5'
            header['EQUINOX'] = 2000.0

    header['MAPTYPE'] = 'FLAT' if flat else 'HEALPIX'
    header['COORDREF'] = str(mask.parent.coord_ref)

    # Append extra metadata in hdr if defined
    if hdr:
        for keyword in hdr.keys():
            header[keyword] = hdr[keyword]

    hdulist = astropy.io.fits.HDUList()
    data = np.asarray(mask)

    if flat:
        if compress:
            ctype = None
            if compress is True:
                ctype = 'GZIP_2'
            elif isinstance(compress, str):
                ctype = compress
            hdu = astropy.io.fits.CompImageHDU(
                data, header=header, compression_type=ctype,
            )
        else:
            hdu = astropy.io.fits.ImageHDU(data, header=header)
    else:
        cols = astropy.io.fits.ColDefs([])
        col = astropy.io.fits.Column(name="MASK", format="B", array=data)
        cols.add_col(col)
        hdu = astropy.io.fits.BinTableHDU.from_columns(cols, header=header)

    hdulist.append(hdu)
    hdulist.writeto(filename, overwrite=overwrite)


@core.indexmod
def SaveMapFrame(
    frame,
    output_file=None,
    hdr=None,
    compress=False,
    quantize_level=16.0,
    overwrite=False,
):
    """
    Save a map frame to a FITS file.  See ``save_skymap_fits`` for details.  The
    map frame should contain T maps and (optionally) unpolarized weights, or T/Q/U
    maps and (optionally) polarized weights to store in the output file.

    Arguments
    ---------
    output_file : str or callable
        Fits filename to which the map will be written.  Maybe a callable
        function that takes a frame object as its sole input argument and
        returns a string filename.
    hdr  : dict
       If defined, extra keywords to be appened to the FITS header. The dict
       can contain entries such as ``hdr['NEWKEY'] = 'New value'`` or
       ``hdr['NEWKEY'] = ('New value', "Comment for New value")``.
    compress : str or bool
        If defined, and if input maps are FlatSkyMap objects, store these in a
        series of compressed image HDUs, one per map.  Otherwise, store input
        maps in a series of standard ImageHDUs, which are readable with older
        FITS readers (e.g. idlastro). If defined, the compression algorithm to
        be used by the Astropy class astropy.io.fits.CompImageHDU.
        Can be: 'RICE_1', 'RICE_ONE', 'PLIO_1', 'GZIP_1', 'GZIP_2' or
        'HCOMPRESS_1'. Only GZIP_1 and GZIP_2 are lossless, although only
        for integer data.
    quantize_level : float
        Floating point quantization level for compression.  Higher values result
        in more accurate floating point representation, but worse compression
        ratio.  See the astropy FITS image documention for details:
        https://docs.astropy.org/en/stable/io/fits/api/images.html
    overwrite : bool
        If True, any existing file with the same name will be ovewritten.
    """

    if frame.type != core.G3FrameType.Map:
        return

    T = frame['T']
    Q = frame.get('Q', None)
    U = frame.get('U', None)
    W = frame.get('Wpol', frame.get('Wunpol', None))

    if callable(output_file):
        output_file = output_file(frame)

    save_skymap_fits(
        output_file,
        T=T,
        Q=Q,
        U=U,
        W=W,
        overwrite=overwrite,
        compress=compress,
        quantize_level=quantize_level,
        hdr=hdr,
    )



#!/usr/bin/env python

from spt3g import core
import time, sys

# File to disk
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
n = 0
def addinfo(fr):
	global n
	if fr.type == core.G3FrameType.EndProcessing:
		return
	fr['time'] = core.G3Time(int(time.time()*core.G3Units.s))
	fr['count'] = n
	n += 1
pipe.Add(addinfo)
pipe.Add(core.Dump)

m = 0
def checkinfo(fr):
	global m
	if fr.type == core.G3FrameType.EndProcessing:
		return
	if 'time' not in fr:
		print('No time key in frame')
		sys.exit(1)
	if fr['count'] != m:
		print('Out of order frame')
		sys.exit(1)
	m += 1
pipe.Add(checkinfo)
pipe.Run()

if m != n:
	print('Wrong number of frames (%d should be %d)' % (m, n))
	sys.exit(1)

sys.exit(0)




from .._libgcp import *

from .ARCExtractor import UnpackACUData, UnpackTrackerData, DecryptFeatureBit, ARCExtract, ARCExtractMinimal
from .ARCHKExtractor import UnpackSPTpolHKData
from .GCPDataTee import GCPHousekeepingTee, GCPSignalledHousekeeping, GCPBoloDataTee, PagerWatchdog, DAQWatchdog
from .InfluxDB import UpdateDB
from .ARCFile import ARCFile



#!/usr/bin/env python

from spt3g import core
from spt3g.core import G3FrameType

frametypes = [G3FrameType.Wiring, G3FrameType.Calibration, G3FrameType.Calibration, G3FrameType.Timepoint, G3FrameType.Observation, G3FrameType.Wiring, G3FrameType.Observation, G3FrameType.Calibration, G3FrameType.Scan, G3FrameType.Scan, G3FrameType.Wiring]

frames = []
for i, frametype in enumerate(frametypes):
	f = core.G3Frame(frametype)
	f['Seq'] = i
	frames.append(f)

out = []

pipe = core.G3Pipeline()

sent = False
def framesource(fr):
	global sent
	if not sent:
		sent = True
		return frames
	else:
		return []
pipe.Add(framesource)

pipe.Add(core.DropOrphanMetadata)

def getframes(fr):
	if fr.type != G3FrameType.EndProcessing and fr.type != G3FrameType.PipelineInfo:
		out.append(fr)
pipe.Add(getframes)

pipe.Run()

assert(len(out) == len(frames) - 3)
def dataframes(frs):
	return [f for f in frs if f.type == core.G3FrameType.Scan or f.type == core.G3FrameType.Timepoint]
assert(len(dataframes(out)) == len(dataframes(frames)))




from . import G3Module, G3FrameObject, usefulfunc
import textwrap


__all__ = ["G3Documenter", "module_apidoc"]


def rst_header(title, line, overline=False):
    """
    Add a valid RST underline (and optional matching overline) to create a
    section header.
    """
    out = title + "\n" + line * len(title) + "\n\n"
    if overline:
        out = line * len(title) + "\n" + out
    return out


@usefulfunc
class G3Documenter:
    """
    Class for inspecting sub-modules of the SPT-3G software package and
    generating valid RST for use with sphinx-autodoc.
    """

    # API categories for which to generate documentation, with a
    # corresponding section title, description and sphinx-autodoc directive.
    _categories = {
        "object": (
            "Frame Objects",
            "Serializable objects that may be stored as entries in "
            ":py:class:`~spt3g.core.G3Frame` objects.",
            "autoclass",
        ),
        "cmodule": (
            "Class-like Pipeline Modules",
            """
            Classes that can be added to :py:class:`~spt3g.core.G3Pipeline`
            instances, either written as callable objects in Python (see
            :ref:`class-modules`) or as :py:class:`~spt3g.core.G3Module`-derived
            classes in C++ (see :ref:`cxx-modules`).  Aside from adding these into
            pipelines using the standard ``Add()`` method, such objects may also be
            instantiated outside of pipelines to make use of additional features.

            One typical use case is to extract some kind of data from a series of
            frames::

                class ExtractData:
                    def __init__(self, key="RawTimestreams"):
                        self.key = key
                        self.data = []
                    def __call__(self, frame):
                        if self.key in frame:
                            self.data.append(frame[self.key])

                # instantiate caching class
                extractor = ExtractData()

                # use instance in a pipeline
                pipe = core.G3Pipeline()
                pipe.Add(core.G3Reader, filename="file.g3")
                pipe.Add(extractor)
                pipe.Run()

                # use extracted data in later processing
                print(extractor.data)

            Another use case may be to call additional methods of a class.  For
            example, the :py:class:`~spt3g.core.G3Writer` class can be used as a
            context manager, and also report the position of the file pointer after
            each frame is written::

                with core.G3Writer("file.g3") as writer:
                    for frame in frames:
                        writer(frame)
                        print("Bytes written:", writer.tell())

            Class-like modules permit a variety of additional features beyond
            their standard usage in pipelines.
            """,
            "autoclass",
        ),
        "fmodule": (
            "Function-like Pipeline Modules",
            "Python functions that can be added to :py:class:`~spt3g.core.G3Pipeline` "
            "instances. Such functions may also be called directly with a "
            ":py:class:`~spt3g.core.G3Frame` object as the first argument, and do not "
            "necessarily need to be used in a pipeline. See :ref:`function-modules` "
            "for more detail.",
            "autofunction",
        ),
        "segment": (
            "Pipeline Segments",
            "Python functions that combine a sequence of pipeline modules, and can be "
            "added to :py:class:`~spt3g.core.G3Pipeline` instances. "
            "See :ref:`pipesegments` for more detail.",
            "autofunction",
        ),
        "class": (
            "Useful Classes",
            "Various Python and C++ classes that are part of the public API.",
            "autoclass",
        ),
        "function": (
            "Useful Functions",
            "Various Python and C++ functions that are part of the public API.",
            "autofunction",
        ),
        "decorator": (
            "Decorators",
            "Decorator functions for indicating any of the above types and/or "
            "modifying function behavior.",
            "autodecorator",
        )
    }

    def __init__(self, root):
        self.root = root
        self.cache = {}

    @property
    def categories(self):
        """
        Dictionary of API categories for which to generate documentation,
        containing a section title, description, and sphinx-autodoc directive as
        a tuple for each category.
        """
        return self._categories

    def find_objects(self, mod=None):
        """
        Recursively find all objects in the input root that fit into one of the
        API categories listed in ``categories``.

        Arguments
        ---------
        mod : object
            Object to inspect

        Returns
        -------
        cache : dict
            Recursively populated dictionary of fully qualified names and their
            associated API category.
        """
        if mod is None:
            mod = __import__(self.root)
            name = self.root.split(".")[-1]
            if '.' in self.root and hasattr(mod, name):
                mod = mod.__dict__[name]

        for x, obj in mod.__dict__.items():
            if x.startswith("_") or x.endswith("_"):
                continue

            # construct fully-qualified object name
            if hasattr(obj, "__wrapped__"):
                obj = obj.__wrapped__
            try:
                if hasattr(obj, '__module__'):
                    itemname = '%s.%s' % (obj.__module__, obj.__name__)
                else:
                    itemname = obj.__name__
            except:
                continue

            # skip objects outside of namespace
            if not itemname.startswith(self.root):
                continue
            # skip already-classified objects
            if itemname in self.cache:
                continue

            # classify
            isclass = isinstance(obj, type)
            ismod = isclass and issubclass(obj, G3Module)
            isobj = isclass and issubclass(obj, G3FrameObject)
            tstr = str(type(obj))
            iscxx = 'builtin_function_or_method' in tstr or 'pybind11' in tstr

            # add to cache
            if hasattr(obj, "__g3decorator__"):
                self.cache[itemname] = "decorator"
            elif ismod or hasattr(obj, "__g3module__"):
                self.cache[itemname] = "cmodule" if isclass else "fmodule"
            elif isobj or hasattr(obj, '__g3frameobject__'):
                self.cache[itemname] = "object"
            elif hasattr(obj, "__pipesegment__"):
                self.cache[itemname] = "segment"
            elif iscxx or hasattr(obj, '__g3usefulfunc__'):
                self.cache[itemname] = "class" if isclass else "function"
            else:
                self.cache[itemname] = "skip"
                if not isclass and hasattr(obj, '__dict__'):
                    # recurse
                    self.find_objects(obj)

        # drop all skipped objects
        if self.root == mod.__name__:
            for k in list(self.cache):
                if self.cache[k] == "skip":
                    self.cache.pop(k)

        return self.cache

    def generate(self):
        """
        Construct valid RST for each of the API types in separate sections, and
        list all relevant objects with appropriate sphinx-autodoc directives.

        Returns
        -------
        str :
            String containing valid RST to append to a document
        """
        # find all relevant objects
        cache = self.find_objects()
        if not len(cache):
            return

        txt = rst_header("API Documentation", "=")

        # build API sections
        body = ""
        for kind, (title, description, directive) in self.categories.items():
            objs = sorted([k for k, v in cache.items() if v == kind])
            if not len(objs):
                continue

            body += rst_header(title, "-")
            body += textwrap.dedent(description) + "\n\n"
            body += "\n\n".join([f".. {directive}:: {k}" for k in objs]) + "\n\n"

        if len(body):
            txt += ".. contents:: Contents\n   :depth: 1\n   :local:\n\n"
            txt += body

        return txt


@usefulfunc
def module_apidoc(module_path):
    """
    Create API documentation for a submodule of the spt3g library.

    The output separates each of the API types into separate sections, and lists
    all relevant objects with appropriate sphinx-autodoc directives.

    Arguments
    ---------
    module_path : str
        Python module path to inspect, e.g. "spt3g.core"

    Returns
    -------
    str :
        String containing valid RST to append to a document
    """

    return G3Documenter(module_path).generate()



"""
Author : Joshua Montgomery <Joshua.Montgomery@mail.mcgill.ca>
Date   : June 16, 2014

This module contains algorithms that measure and calculate quantities related
to the readout, such as SQUID Transimpedance and detector Resistance, Current
and Voltage. It also contains the transfer functions for the ADC and Synthesizer
chains, the SQUID Bias DACs, and the streamers.


for more information, consult the doc-strings
"""

# NOTE: This file copied from pydfmux. Please ensure they stay in sync!

import numpy as np
import json
import functools
import os

# Figure out pydfmux directory and names and fine the json file containing
# the frequency dependent transfer function
TFpath = os.path.dirname(os.path.realpath(__file__))
TFDicts = {}

def frequency_correction(frequency, gain, target='nuller', custom_TF=None):
    """Returns fractional correction to the DC transfer function for a requested
    frequency and analog gain (defaults to baseline gain if none is provided)

    When the transferfunction_utils module is imported the pickle file
    containing the splines with the frequency corrections to the synthesizer
    transfer functions are loaded. This just selects the correct spline from that
    dictionary, evaluates it at the given frequency and returns the correction
    as a float.

    Parameters
    ----------
    frequency : float
        The frequency the correction is requested for (in HZ)

    gain : integer [0, 15]
        The analog gain of the system. The bandwidth changes slightly as
        a function of gain.

    target : string ['nuller', 'carrier'], default='nuller'

    custom_TF : string, default=None
        Optional, allows using a custom frequency correction.  Looks for a file
        in this directory with the argument name

    Returns
    -------
    float
        fractional correction to the transfer function at the provided frequency"""

    try: ## Assume frequency is an iterable
        warn = max(frequency)>9e6
    except TypeError: ## Only given one number
        warn = frequency>9e6

    if custom_TF is None:
        raise ValueError('custom_TF argument is required')

    try:
        tf_model = TFDicts[custom_TF]
    except KeyError:
        TFfile = open('{0}/{1}.txt'.format(TFpath, custom_TF), 'r')
        tf_model = json.load(TFfile)
        TFfile.close()
        TFDicts[custom_TF] = tf_model

    if 'type' in tf_model and tf_model['type'] == 'spline_interpolation':
        from scipy import interpolate

        tf_funct = lambda x: interpolate.splev(x, tf_model[target][str(gain)])
    else:
        tf_funct = functools.partial(np.polyval, tf_model['Gain_{0}'.format(int(gain))][target])
    return tf_funct(frequency)

def gsetting_to_R(G, oldmezz=False):
    """This function converts an analog gain setting into the corresponding effective
    resistance set in the switchable resistor.

    Parameters
    ----------
    G : Integer [0, 15]
        Mezzanine gain setting

    oldmezz : Boolean, default=False
        For use with the Rev2 mezzanines (Not part of any PB2 or SPT3G allocation)

    Returns
    -------
    float
        Effective resistance of the switchable resitive network for the gain
        stage.
    """

    if G not in range(16):
        raise(Exception("Mezzanine Gain settings are integers in set [0, 15]"))

    if oldmezz:
        return (100+(15-G)*15.)
    else:
        Rs = [300.0000, 212.0000, 174.2857, 153.3333, 140.0000, 130.7692,
              124.0000, 118.8235, 114.7368, 111.4286, 108.6957, 106.4000,
              104.4444, 102.7586, 101.2903, 100.0000]
        return Rs[G]

def convert_TF(gain, target, unit='NORMALIZED', frequency=False, custom_TF=None):
    """Analytic Transfer Function.

     Parameters
     ----------
    gain : Integer [0, 15]
        Analog gain setting

    target : string
        Options are contained as attributes of DFMUX boards [d.TARGET.NULLER or
        d.TARGET.CARRIER], or alternatively, 'nuller', or 'carrier'

    units : string ['NORMALIZED', 'RAW'], default='NORMALIZED'
        'NORMALIZED' for programmed or 'RAW' for dan streamer

    frequency : float, default=False
        Optional, will apply a frequency dependent correction if included.

    custom_TF : string, default=None
        Optional, allows using a custom frequency correction.  Looks for a file
        in this directory with the argument name

    Note
    ----
    All amplitudes in the DfMUX system are expressed as Peak Amplitudes.
    This function converts those quantities into real units (Volts and Amps),
    which are still expressed in Peak Amplitude. To convert to RMS, divide
    the resulting values by sqrt(2).

    Returns
    -------
    float
        If target='carrier' the return is Volts Peak Amplitude across Bolometer
        If target='nuller'  the return is Amps Peak Amplitude across the SQUID
    """

    if unit.upper() not in ['RAW', 'NORMALIZED']:
        raise(NameError('Acceptable units are "RAW" or "NORMALIZED"'))
    elif unit.upper()=='RAW':
        prefactor = 1/2.**23
    else:
        prefactor = 1

    if frequency:
        freq_corr = frequency_correction(frequency, gain, target, custom_TF)
    else:
        freq_corr = 1

    if target.lower()=='carrier':
        # The 10 in front is the range of the DAC in mA. the 1e-3 at the end
        # (with the 0.03 mOhm bias resistor) converts this to Volts
        TF_corr = 10*(300*(200/gsetting_to_R(gain))*(1/180.)*(.03))*1e-3
    elif target.lower()=='nuller':
        # The 10 in front is the range of the DAC in mA. the 1e-3 at the end
        # converts this to Amps
        TF_corr = 10*(300. * (200./gsetting_to_R(gain))*(96.77/(100+96.77)) / (750.*4)) * 1e-3
    else:
        raise NameError('Target must be NULLER or CARRIER')
    return prefactor*freq_corr*TF_corr

def convert_adc_samples(unit='ADC Counts'):
    """
    Provides the conversion factor to go from raw ADC samples or Streamer counts
    to a voltage output at the SQUID in Volts.

    Parameters
    ----------
    unit : string ['ADC Counts', 'Streamer'], default='ADC Counts'
        Starting unit. 'ADC Counts' to convert from adc units such as those
        from get_fast_samples. 'Streamer' to convert from the units used in the
        demodulator streamer channels (ledgerman, parser, get_samples, etc)

    Returns
    -------
    float
        Value in volts at the SQUID Output.
    """

    if not unit.upper()=='ADC COUNTS':
        prefactor = 1/128. # This came from a demod streamer -- between the
                           # 16-vs-24 bits scaling and the presence of the
                           # mixer, this is a factor of 128 scaling
    else:
        prefactor = 1.

    gain_1st_stage = 16 # On SQCB (topology: 1+G1/G2)
    gain_2nd_stage = 4  # On SQCB (inverting topology: G1/G2)
    gain_3rd_stage = 4  # On Mezz, with 50ohm on SQCB and 50ohn on Mezz (inverting topology: G1/G2)
    gain_4th_stage = 0.5 # Passive network between last amplifier and ADC chip
    adc_bits = 16 # SQUID has P-P voltage of 2 Volts. AROCs are in Peak Amplitudes

    vsq_per_aroc = ( (1./ (gain_1st_stage*gain_2nd_stage*gain_3rd_stage*gain_4th_stage))
                    /(2**(adc_bits-1)) )

    return vsq_per_aroc*prefactor




#!/usr/bin/env python

from spt3g import core, maps

import numpy

m = maps.HealpixSkyMap(nside=256)

data = numpy.random.uniform(0, 1, size=m.size) > 0.5
mask = maps.G3SkyMapMask(m, data)

fr = core.G3Frame(core.G3FrameType.Map)
fr['mask'] = mask

w = core.G3Writer(filename='masktest.g3')
w(fr)
del w

f = core.G3File('masktest.g3').next()
recovereddata = numpy.asarray(f['mask'])

assert (recovereddata == data).all()




#!/usr/bin/env python

from spt3g import core
import sys

# File to disk
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
n = 0
def addinfo(fr):
	global n
	if fr.type != core.G3FrameType.Timepoint:
		return
	fr['count'] = n
	n += 1
pipe.Add(addinfo)
pipe.Add(lambda frame: 'count' in frame and frame['count'] > 5)
pipe.Add(core.Dump)
ids = []
def getids(fr):
	if fr.type != core.G3FrameType.Timepoint:
		return
	ids.append(fr['count'])
pipe.Add(getids)

pipe.Run()

print(ids)
if ids != [6,7,8,9]:
	sys.exit(1)

sys.exit(0)




from .. import core
from . import DfMuxHousekeepingMap, HkBoardInfo, HkMezzanineInfo, HkModuleInfo, HkChannelInfo, DfMuxWiringMap, DfMuxChannelMapping

from .TuberClient import TuberClient
import socket, struct, time
import numpy

@core.indexmod
class HousekeepingConsumer(object):
    '''
    Collect housekeeping data from the mux boards defined in the wiring map.
    Will add a key called 'DfMuxHousekeeping' to any housekeeping frame that
    goes by containing the data as of the arrival of the housekeeping frame.
    Use in conjunction with a dfmux.PeriodicHousekeepingCollector to get
    data at fixed intervals.

    Also emits a Wiring frame from the housekeeping data.  This requires a
    recent (as of November 2018) version of pydfmux in order to read mapped
    channel names from each board.  If ignore_wiring=False, assumes that
    the wiring map is handled by a separate process.

    If collecting real-time data, you may want to set subprocess=True when
    adding this module.
    '''
    def __init__(self, ignore_wiring=False):
        self.tuber = {}
        self.board_map = {}
        self.board_serials = []
        self.buffer = []
        self.hwmf = None
        self.ignore_wiring = ignore_wiring

    def map_boards(self):
        '''
        Cache IP address, crate serial and slot number and
        create a TuberClient for each IceBoard.
        '''
        boards = [b for b in self.board_serials if b not in self.board_map]
        if not boards:
            return

        ips = {}
        crates = {}
        slots = {}

        # create tubers for each missing board
        for board in boards:
            ip = socket.inet_aton(socket.gethostbyname('iceboard{}.local'.format(board)))
            ips[board] = struct.unpack("i", ip)[0]
            self.tuber[board] = TuberClient(socket.inet_ntoa(ip), timeout=5.0)

        # get crate ID for each missing board
        for board in boards:
            self.tuber[board].CallMethod('Dfmux', '_get_backplane_serial', False)
        for board in boards:
            reply = self.tuber[board].GetReply()[0]
            if reply.get('error', None):
                crates[board] = -1
            else:
                crates[board] = int(reply['result'])

        # get slot ID for each missing board
        for board in boards:
            if crates[board] >= 0:
                self.tuber[board].CallMethod('Dfmux', 'get_backplane_slot', False)
        for board in boards:
            if crates[board] < 0:
                slots[board] = -1
            else:
                reply = self.tuber[board].GetReply()[0]
                if reply.get('error', None):
                    slots[board] = -1
                else:
                    slots[board] = int(reply['result'])

        # store
        for board in boards:
            self.board_map[board] = (ips[board], crates[board], slots[board])

    def __call__(self, frame):

        # If boards have been mapped already, then process every frame
        # as received.
        if len(self.board_map):
            return self.ProcessBuffered(frame)

        # Otherwise, process at least one Timepoint frame first to gather
        # a list of IceBoards for which housekeeping data are required,
        # Then process any buffered frames after Timepoint frame,
        # but emit before the Timepoint
        if frame.type == core.G3FrameType.Timepoint:
            tp_frames = self.ProcessBuffered(frame)
            frames = sum([self.ProcessBuffered(fr) for fr in self.buffer], [])
            frames += tp_frames
            self.buffer = []
            return frames

        # Buffer until a Timepoint frame is received
        self.buffer.append(frame)
        return False

    def ProcessBuffered(self, frame):
        '''
        Process frames in the buffered order.  Returned value should
        always be a list of frames, possibly empty.
        '''
        if frame.type == core.G3FrameType.Timepoint:
            self.board_serials = [('%04d' % k) for k in frame['DfMux'].keys()]
            return [frame]

        if frame.type == core.G3FrameType.Wiring:
            if self.ignore_wiring:
                core.log_warn(
                    "Received a wiring frame, which may be inconsistent with "
                    "board housekeeping data.  Store mapped channel names on the "
                    "board using updated pydfmux/hidfmux software.",
                    unit="HousekeepingConsumer",
                )
            else:
                core.log_fatal(
                    "Received spurious wiring frame.  Do not use "
                    "PyDfMuxHardwareMapInjector with the HousekeepingConsumer.  "
                    "You may update pydfmux/hidfmux to a newer version that "
                    "stores mapped channel names on the boards, and rearrange "
                    "your data acquisition script.",
                    unit="HousekeepingConsumer",
                )

        if frame.type == core.G3FrameType.Housekeeping:
            self.map_boards()

            hwm = DfMuxWiringMap()
            hkdata = DfMuxHousekeepingMap()
            try:
                for board in self.board_serials:
                    self.tuber[board].CallMethod('Dfmux', '_dump_housekeeping', False)
                    time.sleep(0.02) # Stagger return data transfer a little to
                                     # avoid overloading the network on the return

                found = False
                ismkid = False
                for board in self.board_serials:
                    dat = self.tuber[board].GetReply()[0]['result']

                    boardhk = self.HousekeepingFromJSON(dat)
                    if boardhk.firmware_name:
                        ismkid = ismkid or "mkid" in boardhk.firmware_name.lower()
                    hkdata[int(boardhk.serial)] = boardhk

                    if self.ignore_wiring:
                        continue

                    ip, crate, slot = self.board_map[board]
                    boardw = self.WiringFromJSON(dat, ip, crate, slot)
                    for key in boardw.keys():
                        hwm[key] = boardw[key]
                        found = True

                if not found and not self.ignore_wiring:
                    core.log_fatal("No mapped channels found on any IceBoards. "
                                   "You may need to update pydfmux to a newer version of pydfmux "
                                   "that stores mapped channel names on the boards, and reload "
                                   "the hardware map.",
                                   unit='HousekeepingConsumer')

                frame['DfMuxHousekeeping'] = hkdata

                if self.ignore_wiring:
                    return [frame]

                hwmf = core.G3Frame(core.G3FrameType.Wiring)
                hwmf['WiringMap'] = hwm
                hwmf['ReadoutSystem'] = 'RF-ICE' if ismkid else 'ICE'

                if self.hwmf is None:
                    self.hwmf = hwmf
                    # If this is the first time the consumer is triggered, make sure
                    # a Housekeeping and WiringMap frame are issued before any
                    # Timepoint frames
                    frames = [hwmf, frame]
                    return frames
                else:
                    # Compare wiring maps and re-issue frame if anything has changed
                    old_hwm = self.hwmf['WiringMap']
                    if (set(hwm.keys()) ^ set(old_hwm.keys())):
                        self.hwmf = hwmf
                        return [hwmf, frame]
                    for k in hwm.keys():
                        try:
                            if vars(hwm[str(k)]) != vars(old_hwm[str(k)]):
                                self.hwmf = hwmf
                                return [hwmf, frame]
                        except:
                            core.log_error("Invalid HWM key %r" % k)

                # If we get here then the wiring map hasn't changed,
                # so return the populated Housekeeping frame as it is
                return [frame]

            except socket.timeout:
                core.log_error('Timeout collecting housekeeping data from mux boards. Dropping housekeeping sample', unit='HousekeepingConsumer')
                return []
            except Exception as e:
                core.log_error('Error (%s) collecting housekeeping data from mux boards. Dropping housekeeping sample' % e, unit='HousekeepingConsumer')
                return []

        return [frame]

    @classmethod
    def HousekeepingFromJSON(cls, dat):
        '''
        Build HKBoardInfo object from a JSON blob returned by the
        _dump_housekeeping call
        '''
        # Board-global quantities
        boardhk = HkBoardInfo()
        if 'is128x' in dat:
            boardhk.is128x = dat['is128x']
        else:
            boardhk.is128x = False
        year = dat['timestamp']['y']
        if year == 0:
            # It probably isn't 1900
            systime = time.gmtime()
            # Check for New Year's, assuming no more than 24 hours clock slew
            if dat['timestamp']['d'] == 1 and systime.tm_yday >= 365:
                year = systime.tm_year + 1
            elif dat['timestamp']['d'] >= 365 and systime.tm_yday == 1:
                year = systime.tm_year - 1
            else:
                year = systime.tm_year
            year -= 2000
        boardhk.timestamp = core.G3Time(y=year,d=dat['timestamp']['d'],h=dat['timestamp']['h'],m=dat['timestamp']['m'],s=dat['timestamp']['s'],ss=dat['timestamp']['ss'])

        boardhk.timestamp_port = str(dat['timestamp_port'])
        boardhk.serial = str(dat['serial'])
        if 'firmware_name' in dat:
            boardhk.firmware_name = str(dat['firmware_name'])
            boardhk.firmware_version = str(dat['firmware_version'])
        boardhk.fir_stage = dat['fir_stage']
        for i in dat['currents'].items():
            boardhk.currents[str(i[0])] = i[1]
        for i in dat['voltages'].items():
            boardhk.voltages[str(i[0])] = i[1]
        for i in dat['temperatures'].items():
            boardhk.temperatures[str(i[0])] = i[1]

        # Mezzanines
        for n, mezz in enumerate(dat['mezzanines']):
            mezzhk = HkMezzanineInfo()
            mezzhk.present = mezz['present']
            mezzhk.power = mezz['power']
            if mezzhk.present:
                if 'ipmi' in mezz:
                    mezzhk.serial = str(mezz['ipmi']['product']['serial_number'])
                    mezzhk.part_number = str(mezz['ipmi']['product']['part_number'])
                    mezzhk.revision = str(mezz['ipmi']['product']['version_number'])
                if 'currents' in mezz:
                    for i in mezz['currents'].items():
                        mezzhk.currents[str(i[0])] = i[1]
                if 'voltages' in mezz:
                    for i in mezz['voltages'].items():
                        mezzhk.voltages[str(i[0])] = i[1]

            if mezzhk.present and mezzhk.power:
                if 'temperature' in mezz:
                    mezzhk.temperature = mezz['temperature']
                # these parameters are not in the 64x housekeeping tuber
                mezzhk.squid_heater = mezz.get('squid_heater', 0.0)
                mezzhk.squid_controller_power = mezz.get('squid_controller_power', False)
                mezzhk.squid_controller_temperature = mezz.get('squid_controller_temperature', 0.0)

            # Modules
            for m, mod in enumerate(mezz['modules']):
                modhk = HkModuleInfo()
                modhk.routing_type = str(mod['routing'][0])
                modhk.module_number = m+1
                if mezzhk.present and mezzhk.power:
                    if 'gains' in mod:
                        modhk.carrier_gain = mod['gains']['carrier']
                        modhk.nuller_gain = mod['gains']['nuller']
                    if 'overload' in mod:
                        modhk.carrier_railed = mod['overload']['carrier']
                        modhk.nuller_railed = mod['overload']['nuller']
                        modhk.demod_railed = mod['overload']['demod']
                    if 'squid_current_bias' in mod:
                        modhk.squid_current_bias = mod['squid_current_bias']
                    if 'squid_flux_bias' in mod:
                        modhk.squid_current_bias = mod['squid_flux_bias']
                    if 'squid_feedback' in mod:
                        modhk.squid_feedback = str(mod['squid_feedback'])
                    if 'nco_frequency' in mod:
                        modhk.nco_frequency = mod['nco_frequency'] * core.G3Units.Hz

                if 'squid_tuning' in mod and mod['squid_tuning'] is not None:
                    modhk.squid_state = str(mod['squid_tuning']['state'])
                    modhk.squid_transimpedance = mod['squid_tuning']['transimpedance'] if mod['squid_tuning']['transimpedance'] is not None else numpy.nan
                    modhk.squid_p2p = mod['squid_tuning']['p2p'] if mod['squid_tuning']['p2p'] is not None else numpy.nan

                for k, chan in enumerate(mod['channels']):
                    chanhk = HkChannelInfo()
                    chanhk.channel_number = k+1
                    chanhk.carrier_amplitude = chan['carrier_amplitude']
                    chanhk.nuller_amplitude = chan['nuller_amplitude']
                    if 'dan_gain' in chan:
                        chanhk.dan_gain = chan['dan_gain']
                    if 'dan_streaming_enable' in chan:
                        chanhk.dan_streaming_enable = chan['dan_streaming_enable']
                    if 'frequency' in chan:
                        chanhk.carrier_frequency = chan['frequency'] * core.G3Units.Hz
                        chanhk.demod_frequency = chan['frequency'] * core.G3Units.Hz
                    else:
                        chanhk.carrier_frequency = chan['carrier_frequency'] * core.G3Units.Hz
                        chanhk.demod_frequency = chan['demod_frequency'] * core.G3Units.Hz
                    if 'carrier_phase' in chan:
                        chanhk.carrier_phase = chan['carrier_phase'] * core.G3Units.deg
                        chanhk.nuller_phase = chan['nuller_phase'] * core.G3Units.deg
                        chanhk.demod_phase = chan['demod_phase'] * core.G3Units.deg
                    if 'dan_accumulator_enable' in chan:
                        chanhk.dan_accumulator_enable = chan['dan_accumulator_enable']
                    if 'dan_feedback_enable' in chan:
                        chanhk.dan_feedback_enable = chan['dan_feedback_enable']
                    if 'dan_railed' in chan:
                        chanhk.dan_railed = chan['dan_railed']
                    if 'tuning' in chan and chan['tuning'] is not None:
                        if "tune" in chan["tuning"] and "state" not in chan["tuning"]:
                            # KIDs
                            chanhk.state = "tuned" if chan["tuning"]["tune"] else "unknown"
                        else:
                            chanhk.state = str(chan["tuning"]["state"])
                        attrs = {
                            "rlatched": ("rlatched", core.G3Units.ohm),
                            "rnormal": ("rnormal", core.G3Units.ohm),
                            "rfrac_achieved": ("rfrac_achieved", 1),
                            "loopgain": ("loopgain", 1),
                            "i_slope": ("dIr", 1. / core.G3Units.Hz),
                            "q_slope": ("dQr", 1. / core.G3Units.Hz),
                            "internal_phase": ("internal_phase", core.G3Units.rad),
                            "external_phase": ("external_phase", core.G3Units.rad),
                            "bias_frequency": ("bias_freq", core.G3Units.Hz),
                        }
                        for attr, (na, ua) in attrs.items():
                            if chan["tuning"].get(na, None) is not None:
                                setattr(chanhk, attr, float(chan["tuning"][na]) * ua)

                    modhk.channels[k+1] = chanhk
                mezzhk.modules[m+1] = modhk
            boardhk.mezz[n+1] = mezzhk
    
        return boardhk

    @classmethod
    def WiringFromJSON(cls, dat, ip, crate, slot):
        '''
        Build WiringMap object from a JSON blob returned by the
        _dump_housekeeping call
        '''
        found = False
        hwm = DfMuxWiringMap()
        serial = int(dat['serial'])
        for imezz, mezz in enumerate(dat['mezzanines']):
            for imod, mod in enumerate(mezz['modules']):
                module = imod + len(mezz['modules']) * imezz
                for ichan, chan in enumerate(mod['channels']):
                    name = (chan.get('tuning', {}) or {}).get('name', None)
                    if not name:
                        continue
                    mapping = DfMuxChannelMapping()
                    mapping.board_ip = ip
                    mapping.board_serial = serial
                    mapping.board_slot = slot
                    mapping.crate_serial = crate
                    mapping.module = module
                    mapping.channel = ichan
                    try:
                        name = str(name)
                        hwm[name] = mapping
                        found = True
                    except:
                        core.log_error("Invalid channel name %r" % (name))
        if not found:
            core.log_error("No mapped channels found on iceboard%04d. "
                           "You may need to update pydfmux to a newer version of pydfmux "
                           "that stores mapped channel names on the boards, and reload "
                           "the hardware map." % (serial),
                           unit='HousekeepingConsumer')
        return hwm

    @classmethod
    def HousekeepingForBoard(cls, hostname):
        '''
        Return HkBoardInfo object from a board. Useful for debugging.
        '''
        t = TuberClient(hostname)
        data = t.CallMethod('Dfmux', '_dump_housekeeping')
        return cls.HousekeepingFromJSON(data[0]['result'])

@core.indexmod
class PeriodicHousekeepingCollector(object):
    '''Inserts housekeeping frames every N timepoints.'''
    def __init__(self, N=15200):
        self.N = N
        self.count = 0
    def __call__(self, frame):
        ret = []
        if frame.type == core.G3FrameType.Timepoint:
            if self.count % self.N == 0:
                ret.append(core.G3Frame(core.G3FrameType.Housekeeping))
            self.count += 1
            ret.append(frame)
            return ret

@core.usefulfunc
def HousekeepingForBolo(hkmap, wiringmap, bolo, all_hk=False):
    '''
    Obtain the channel housekeeping information for a bolometer named "bolo"
    using the passed housekeeping and wiring maps.

    If all_hk is True, returns a tuple of the (board, mezz, module, channel)
    HK data instead of just the channel.
    '''

    wiringentry = wiringmap[bolo]
    board = hkmap[wiringentry.board_serial]
    mezz = board.mezz[(wiringentry.module // len(board.mezz[1].modules)) + 1]
    mod = mezz.modules[(wiringentry.module % len(mezz.modules)) + 1]
    chan = mod.channels[wiringentry.channel + 1]
    if all_hk:
        return (board, mezz, mod, chan)
    else:
        return chan



#!/usr/bin/env python
from spt3g import core, maps
import numpy as np

a = core.Quat(2,3,4,5)
a2 = core.Quat(2,1,8,5)
b = core.G3VectorQuat([a, a**2, 2*a, a2])

assert(np.allclose(maps.quat_to_ang(a), maps.c_quat_to_ang_(a)))
assert(np.allclose(maps.quat_to_ang(b), np.asarray([maps.c_quat_to_ang_(x) for x in b]).transpose()))

angle = (.3, 0.4)
angles = ((.3, 0.8), (0.4, 0.2))

q = maps.ang_to_quat(angle[0], angle[1])
p = maps.c_ang_to_quat_(angle[0], angle[1])
assert(np.allclose(core.G3VectorQuat([q]), core.G3VectorQuat([p])))
alpha, delta = list(zip(angles[0], angles[1]))
assert(np.allclose(maps.ang_to_quat(alpha, delta), core.G3VectorQuat([maps.c_ang_to_quat_(a[0], a[1]) for a in angles])))

assert(np.allclose(maps.quat_to_ang(maps.c_ang_to_quat_(angle[0], angle[1])), angle))

start = core.G3Time(0)
stop = core.G3Time(10)
q_timestream = maps.ang_to_quat(alpha, delta, start, stop)
assert q_timestream.start == start
assert q_timestream.stop == stop



#!/usr/bin/env python

from spt3g import core
import time, sys

# File to disk
n = 0
def addinfo(fr):
	global n
	if fr.type != core.G3FrameType.Timepoint:
		return
	fr['time'] = core.G3Time(int(time.time()*core.G3Units.s))
	fr['count'] = n
	n += 1
m = 0
def checkinfo(fr):
	global m
	if fr.type != core.G3FrameType.Timepoint:
		return
	if 'time' not in fr:
		raise RuntimeError('No time key in frame')
	if fr['count'] != m:
		raise RuntimeError('Out of order frame')
	m += 1

if __name__ == '__main__':
	pipe = core.G3Pipeline()
	pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
	pipe.Add(addinfo, subprocess=True)
	pipe.Add(core.Dump)
	pipe.Add(checkinfo)
	pipe.Run()

	if n != 0:
		raise RuntimeError('Ran in same process!')
	n = 10 # Different process, so don't get this back
	if m != n:
		raise RuntimeError('Wrong number of frames (%d should be %d)' % (m, n))



import struct, socket, errno, numpy, time, threading
from .. import core, dfmux


@core.indexmod
class PagerWatchdog(object):
    '''
    Module that sends a watchdog (ping) message to the GCP pager when the parent
    process is running successfully.  Modify the `data_valid` method for
    particular use cases, and call the `run` method periodically in your
    application.
    '''

    host = 'sptnet.spt'
    port = 50040
    timeout = 20

    def __init__(self, name, interval=600, sim=False):
        self.name = name.lower()
        self.unit = '{}Watchdog'.format(name.capitalize())
        self.interval = interval
        self.sim = sim
        self.last_ping = None

        # ping on startup
        self.thread = threading.Thread(target=self.ping)
        self.thread.start()

    def send(self, cmd):
        """
        Send a command to the GCP pager process, and return the response.
        """
        if self.sim:
            return

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(self.timeout)
            sock.connect((self.host, self.port))
            sock.send(cmd.encode())
            resp = sock.recv(4096)
            if resp:
                core.log_debug(
                    "Sent command, got response {}".format(resp), unit=self.unit,
                )
            sock.close()
        except Exception as e:
            core.log_error("Error sending command: {}".format(e), unit=self.unit)
            raise
        else:
            core.log_info("Sent command {}".format(cmd), unit=self.unit)

        return resp.decode()

    def ping(self):
        """
        Send a watchdog ping message to the GCP pager process.  This method is
        called by the `run` method at regular intervals whenever the
        `data_valid` method returns True.
        """
        try:
            self.send("watchdog {}".format(self.name))
        except Exception:
            # try again in ten seconds
            self.last_ping = time.time() - self.interval + 10
        else:
            self.last_ping = time.time()

    def page(self, msg):
        """
        Issue a page to the GCP pager process.  Called by the user for immediate
        notification of errors.
        """
        try:
            self.send("page '{}'".format(self.name, msg))
        except Exception as e:
            pass

    def data_valid(self, *args, **kwargs):
        """
        Returns True if the watchdog should ping, otherwise False.  Arguments
        are passed to this method from the `run` method.
        """
        raise NotImplementedError

    def run(self, *args, **kwargs):
        """
        When called, issues a watchdog ping message if the interval time has passed, and
        the `data_valid` method returns True.  All input arguments are passed to the
        `data_valid` method for validation.
        """

        # only ping if ready
        if not self.data_valid(*args, **kwargs):
            return

        # only ping if another ping isn't already running
        if self.thread is not None:
            if not self.thread.is_alive():
                del self.thread
                self.thread = None
            else:
                return

        # only ping on the appropriate interval
        now = time.time()
        if self.last_ping and (now - self.last_ping < self.interval):
            return

        # ping
        self.thread = threading.Thread(target=self.ping)
        self.thread.start()


@core.indexmod
class DAQWatchdog(PagerWatchdog):
    """
    Watchdog that issues a ping to the GCP pager when the DAQ is operational.
    """

    def __init__(self, calibrator=False, interval=600, sim=False):
        """
        Arguments
        ---------
        calibrator : bool
            If True, ensure that the calibrator is also running successfully
            before sending a ping.
        """
        super(DAQWatchdog, self).__init__('DAQ', interval=interval, sim=sim)

        self.last_missing = None
        self.boards_missing = 0
        self.last_log_boards = None
        self.calibrator = calibrator
        self.last_log_cal = None

    def data_valid(self, frame):
        """
        Check the incoming frame for completeness.

         * Ensure that all modules in the listed iceboards are reporting.
         * If `calibrator` is True, ensure that the calibrator sync signal is in the frame.
        """

        # always ready in sim mode
        if self.sim:
            return True

        # only ping on Timepoint frames
        if 'DfMux' not in frame:
            return False

        now = time.time()
        retval = True

        # only ping if all expected modules are present
        data = frame['DfMux']
        boards_expected = len(data.keys())
        boards_complete = sum([v.nmodules > 0 and v.Complete() for v in data.values()])
        boards_missing = boards_expected - boards_complete
        if boards_missing:
            if not self.last_log_boards or boards_missing != self.boards_missing:
                # log loss or change in missing count
                core.log_error(
                    "Missing data from {} boards in DAQ data stream".format(boards_missing),
                    unit=self.unit,
                )
                self.last_log_boards = now
                self.boards_missing = boards_missing
            self.last_missing = now
            retval = False
        elif self.last_log_boards and now - self.last_missing > 10:
            # log recovery
            core.log_notice("All boards recovered in DAQ data stream", unit=self.unit)
            self.last_log_boards = None

        # only ping if the calibrator sync signal is present
        if self.calibrator and 'CalibratorOn' not in frame:
            if not self.last_log_cal:
                # log loss
                core.log_error(
                    "Missing calibrator signal in DAQ data stream",
                    unit=self.unit,
                )
                self.last_log_cal = now
            self.last_missing = now
            retval = False
        elif self.calibrator and self.last_log_cal and now - self.last_missing > 10:
            # log recovery
            core.log_notice(
                "Calibrator signal recovered in DAQ data stream", unit=self.unit
            )
            self.last_log_cal = None

        # only ping if normal data acquisition has been going for a bit
        if self.last_missing and now - self.last_missing < 10:
            retval = False

        return retval

    def __call__(self, frame):
        self.run(frame)


@core.indexmod
class GCPSignalledHousekeeping(object):
    '''
    Module that collects housekeeping data when connected to. If
    collect_on_start is True (the default), injects an HK frame
    unconditionally at startup.
    '''
    def __init__(self, port=50011, collect_on_start=True):
        self.socket = socket.socket()
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.socket.bind(('', port))
        self.socket.listen(5)
        self.socket.setblocking(False)

        self.collect_on_start = collect_on_start
        self.first_frame = True

    def __del__(self):
        self.socket.close()

    def __call__(self, frame):
        # Only try on timepoint frames
        if frame.type != core.G3FrameType.Timepoint:
            return

        if self.collect_on_start and self.first_frame:
            self.first_frame = False
            return [core.G3Frame(core.G3FrameType.Housekeeping), frame]

        # Check for new connections
        try:
            s, origin_ip = self.socket.accept()
        except socket.error as e:
            if e.errno != errno.EAGAIN and e.errno != errno.EWOULDBLOCK:
                raise
            return
        core.log_debug('Accepted housekeeping collection signal from %s:%d' % origin_ip,
                       unit='GCPSignalledHousekeeping')
        s.close()

        return [core.G3Frame(core.G3FrameType.Housekeeping), frame]

@core.indexmod
class GCPHousekeepingTee(object):
    '''
    Module that serves housekeeping information to GCP when asked. If a key
    named "DataOK" exists in the housekeeping frames, will also transmit
    that information to GCP for paging purposes.
    '''
    def __init__(self, port=50010, verbose=False):

        # make some noise at startup
        core.log_info("Initialize gcp.GCPHousekeepingTee on port %d" % port,
                      unit='GCPHousekeepingTee')

        self.hkblob = self.PackHKToGCP(dfmux.DfMuxHousekeepingMap())
        self.socket = socket.socket()
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.socket.bind(('', port))
        self.socket.listen(25)
        self.socket.setblocking(False)

        self.verbose = verbose  # flag for printing debugging statements

    def __del__(self):
        # Clear any pending connections. No one is getting anything now.
        # This works around some misfeatures in the Linux kernel that
        # do not occur in other, better socket implementations.
        while True:
            try:
                s, origin_ip = self.socket.accept()
            except socket.error as e:
                if e.errno != errno.EAGAIN and e.errno != errno.EWOULDBLOCK:
                    raise
                break

            s.close()
        self.socket.close()

    def __call__(self, frame):
        if frame.type == core.G3FrameType.Housekeeping:
            dataok = True
            if 'DataOK' in frame:
                dataok = frame['DataOK']
            self.hkblob = self.PackHKToGCP(
                frame['DfMuxHousekeeping'], dataok=dataok,
                verbose=self.verbose)

            # Check for new connections, send any interested
            # parties the same data
            cxs = []
            while True:
                try:
                    s, origin_ip = self.socket.accept()
                except socket.error as e:
                    if e.errno != errno.EAGAIN and \
                       e.errno != errno.EWOULDBLOCK:
                        raise
                    break

                core.log_debug('Accepted connection from %s:%d' % origin_ip,
                               unit='GCPHousekeepingTee')
                cxs.append(s)

            for s in cxs:
                s.setblocking(True)
                s.sendall(self.hkblob)
                s.close()


    @staticmethod
    def PackHKToGCP(hk, dataok=True, verbose=False):
        if verbose:
            core.log_debug('gcp.GCPHousekeepingTee.PackHKToGCP(hk)', unit='GCPHousekeepingTee')
        buf = struct.pack('<?I', dataok, len(hk))

        # See HkDataStruct in GCP
        for ip, board in hk.items():

            # if verbose mode, print a few registers for debugging
            if verbose:
                core.log_debug("%d, %10.6f, %10.6f, %10.6f, %10.6f, %10.6f" %
                               (ip, board.temperatures['MOTHERBOARD_TEMPERATURE_ARM'],
                                board.temperatures['MOTHERBOARD_TEMPERATURE_FPGA'],
                                board.temperatures['MOTHERBOARD_TEMPERATURE_FPGA_DIE'],
                                board.temperatures['MOTHERBOARD_TEMPERATURE_PHY'],
                                board.temperatures['MOTHERBOARD_TEMPERATURE_POWER']),
                               unit='GCPHousekeepingTee')

            buf += struct.pack('<fffff',
                               board.temperatures['MOTHERBOARD_TEMPERATURE_ARM'],
                               board.temperatures['MOTHERBOARD_TEMPERATURE_FPGA'],
                               board.temperatures['MOTHERBOARD_TEMPERATURE_FPGA_DIE'],
                               board.temperatures['MOTHERBOARD_TEMPERATURE_PHY'],
                               board.temperatures['MOTHERBOARD_TEMPERATURE_POWER'])
            buf += struct.pack('<fffffffff',
                               board.voltages['MOTHERBOARD_RAIL_VADJ'],
                               board.voltages['MOTHERBOARD_RAIL_VCC12V0'],
                               board.voltages['MOTHERBOARD_RAIL_VCC1V0'],
                               board.voltages['MOTHERBOARD_RAIL_VCC1V0_GTX'],
                               board.voltages['MOTHERBOARD_RAIL_VCC1V2'],
                               board.voltages['MOTHERBOARD_RAIL_VCC1V5'],
                               board.voltages['MOTHERBOARD_RAIL_VCC1V8'],
                               board.voltages['MOTHERBOARD_RAIL_VCC3V3'],
                               board.voltages['MOTHERBOARD_RAIL_VCC5V5'])
            buf += struct.pack('<fffffffff',
                               board.currents['MOTHERBOARD_RAIL_VADJ'],
                               board.currents['MOTHERBOARD_RAIL_VCC12V0'],
                               board.currents['MOTHERBOARD_RAIL_VCC1V0'],
                               board.currents['MOTHERBOARD_RAIL_VCC1V0_GTX'],
                               board.currents['MOTHERBOARD_RAIL_VCC1V2'],
                               board.currents['MOTHERBOARD_RAIL_VCC1V5'],
                               board.currents['MOTHERBOARD_RAIL_VCC1V8'],
                               board.currents['MOTHERBOARD_RAIL_VCC3V3'],
                               board.currents['MOTHERBOARD_RAIL_VCC5V5'])
            buf += struct.pack('255s', ('iceboard' + board.serial).encode())
            for i in [1,2]:
                buf += struct.pack('<?',
                                   board.mezz[i].present)
                buf += struct.pack('<?',
                                   board.mezz[i].power)
                buf += struct.pack('<?',
                                   board.mezz[i].squid_controller_power)
                buf += struct.pack('<fff',
                                   board.mezz[i].voltages['MEZZANINE_RAIL_VADJ'],
                                   board.mezz[i].voltages['MEZZANINE_RAIL_VCC12V0'],
                                   board.mezz[i].voltages['MEZZANINE_RAIL_VCC3V3'])
                buf += struct.pack('<fff',
                                   board.mezz[i].currents['MEZZANINE_RAIL_VADJ'],
                                   board.mezz[i].currents['MEZZANINE_RAIL_VCC12V0'],
                                   board.mezz[i].currents['MEZZANINE_RAIL_VCC3V3'])
                buf += struct.pack('<fff',
                                   board.mezz[i].temperature,
                                   board.mezz[i].squid_controller_temperature,
                                   board.mezz[i].squid_heater)


        # Prefix with total message length
        buf = struct.pack('!q', len(buf)) + buf

        return buf

@core.indexmod
class GCPBoloDataTee(object):
    '''
    Module that serves bolometer data to GCP when asked. Once a second,
    will serve the data from the previous second of bolometer data.

    If a boolean key appears in the timepoint frames named "DataOK",
    this will be sent to GCP as a data quality indicator for paging.
    '''
    def __init__(self, port=50020, verbose=False, bolometers=[]):
        '''
        Send data from the given list of bolometer logical IDs to the GCP.
        '''

        core.log_info('Listening for requests from GCP on port %d' % port, unit='GCPBoloDataTee')
        core.log_info('Selected bolometers: %s' % bolometers, unit='GCPBoloDataTee')

        self.bololist = bolometers

        self.socket = socket.socket()
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.socket.bind(('', port))
        self.socket.listen(5)
        self.socket.setblocking(False)
        self.data = {}

    def __call__(self, frame):
        if frame.type == core.G3FrameType.Wiring:
            self.wmap = frame['WiringMap']

        if frame.type == core.G3FrameType.Timepoint:
            sec = int(frame['EventHeader'].time/core.G3Units.s)

            # Add data from this sample to the cache for this calendar second, replacing any missing detectors with -1
            if sec not in self.data:
                self.data[sec] = {b: [] for b in self.bololist}
                self.data[sec]['DataOK'] = []
            d = self.data[sec]
            for b in self.bololist:
                w = self.wmap[b]
                board = frame['DfMux'][w.board_serial]
                if board.nblocks > 1:
                    mod_idx = w.module * board.nblocks + w.channel // board.nchannels
                    chan_idx = w.channel % board.nchannels
                else:
                    mod_idx = w.module
                    chan_idx = w.channel
                try:
                    d[b].append(board[mod_idx][chan_idx])
                except KeyError:
                    d[b].append(-1)
            if 'DataOK' in frame:
                self.data[sec]['DataOK'].append(bool(frame['DataOK']))
            else:
                self.data[sec]['DataOK'].append(True)

            # Toss ancient data: we keep the last second (complete)
            # for GCP, plus the second we are currently accumulating
            if len(self.data) > 2:
                keys = list(self.data.keys())
                keys.sort()
                for k in keys[:-2]:
                    del self.data[k]

        # Check for new connections once we have a buffer
        if len(self.data) == 2:
            try:
                s, origin_ip = self.socket.accept()
            except socket.error as e:
                if e.errno != errno.EAGAIN and e.errno != errno.EWOULDBLOCK:
                    raise
                return

            core.log_debug('Accepted connection from %s:%d' % origin_ip, unit='GCPBoloDataTee')
            s.setblocking(True)
            keys = list(self.data.keys())
            keys.sort()
            s.sendall(self.PackForGCP(self.data[keys[0]]))
            s.close()

            # Delete data once enqueued
            del self.data[keys[0]]


    @staticmethod
    def PackForGCP(data):
        # Input data: dict of bolo names to samples (can be a
        # G3TimestreamMap, in principle)
        #
        # On-wire format:
        # U64 Length of buffer (big-endian)
        # U8 Data Quality Indicator: True (1) = good
        # U32 Number of detectors in list
        # U32 Number of samples in the last second
        # N copies of:
        # - 16 byte character string with detector name
        # - N_sample 32-bit signed integers with data
        #
        # All fields are little-endian, unless otherwise noted

        data_len = len(data[list(data.keys())[0]])
        buf = struct.pack('<?II', numpy.all(data['DataOK']), len(data) - 1, data_len)
        for k, v in data.items():
            if k == 'DataOK':
                continue
            buf += struct.pack('16s', k.encode())
            assert(len(v) == data_len)
            buf += struct.pack('<%di' % data_len, *v)

        buf = struct.pack('!q', len(buf)) + buf

        return buf




from sphinx.ext import autodoc
import pprint
import re
import textwrap
import spt3g

# sphinx configuration options

project = "SPT-3G Software"
version = spt3g.__version__
copyright = "2017-, SPT-3G Collaboration"
author = "SPT-3G Collaboration"
master_doc = "index"
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    "collapse_navigation": False,
    "navigation_depth": 4,
}
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_logo = "spt_outlines.png"
html_favicon = "spt_outlines_white.png"
exclude_patterns = ["intro_*"]
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
]

# docstring parser configuration
napoleon_custom_sections = [
    ("Emits", "returns_style"),
    ("Equivalent to", "Example")
]
maximum_signature_line_length = 85
python_use_unqualified_type_names = True

# autodoc configuration
autodoc_class_signature = "mixed"
autoclass_content = "both"
autodoc_member_order = "groupwise"
autodoc_typehints = "description"
autodoc_typehints_format = "short"
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
    "class-doc-from": "both",
    "member-order": "groupwise",
    "exclude-members": "__new__",
}

# custom documenters for handling C++ signatures and overloaded functions

selfarg = re.compile(r"(\(self\:\s*[\w.]+(?:,\s*)?)", re.VERBOSE)
arg0 = re.compile(r"\(arg0:(\s*[\w.]+)\)", re.VERBOSE)


class BindingDocsMixin:
    """
    Mixin class to handle signature cleanup in C++ bindings,
    including annotations and overloaded signatures
    """
    def get_doc(self):
        if self._new_docstrings is not None:
            return self._new_docstrings

        docstrings = super().get_doc()
        if not docstrings:
            return docstrings

        tstr = str(type(self.object))
        if (
            'pybind11' not in tstr
            and 'builtin_function_or_method' not in tstr
            and 'instancemethod' not in tstr
        ):
            return docstrings

        for i, doclines in enumerate(docstrings):
            overloaded = False
            if "Overloaded function." in doclines:
                overloaded = True
                while doclines[0].strip() != "Overloaded function.":
                    doclines.pop(0)
                doclines.pop(0)

            # drop empty lines
            while not doclines[0].strip():
                doclines.pop(0)

            # collect all signature lines
            sigs = []
            docs = []
            docs1 = []
            for j, line in enumerate(doclines):
                # CXX bindings typically have return annotations
                if (not overloaded and j > 0) or (overloaded and "->" not in line):
                    docs1.append(line)
                    continue

                # remove unnecessary return annotations
                if overloaded and re.match(r"^[0-9]\.\s+", line):
                    line = line.split(".", 1)[-1].lstrip()
                if line.startswith("__init__"):
                    line = line.partition(" -> ")[0]
                if line.endswith("-> None"):
                    line = line.replace(" -> None", "")
                # remove self from method argument list
                m = selfarg.search(line)
                if m:
                    line = line.replace(m.group(0), "(")
                # rename single arg0 to arg
                m = arg0.search(line)
                if m:
                    line = line.replace(m.group(0), "(arg:" + m.group(1) + ")")
                line = line.replace("()", "(\u200b)")
                line = re.sub('<(.*) at (.*)>', '<\\1>', str(line))

                # collate overloaded signatures and corresponding docstrings
                if line.strip():
                    if "".join(docs1).strip():
                        if sigs:
                            n = len(docs1[0]) - len(docs1[0].lstrip())
                            docs1[0] = docs1[0][:n] + "Signature {}: {}".format(
                                len(sigs), docs1[0].lstrip()
                            )
                        docs.extend(docs1 + [""])
                    sigs.append(line)
                    docs1.clear()
                doclines[j] = line

            if len(sigs):
                # finish collation
                doclines.clear()
                doclines.extend(sigs)
                if "".join(docs1).strip():
                    if len(sigs) > 1 and (len(docs) or i > 0):
                        n = len(docs1[0]) - len(docs1[0].lstrip())
                        docs1[0] = docs1[0][:n] + "Signature {}: {}".format(
                            len(sigs), docs1[0].lstrip()
                        )
                    docs.extend(docs1 + [""])
                doclines.extend(docs)

        return docstrings


# Create mixin subclasses for all relevant documenters
class FunctionDocumenter(BindingDocsMixin, autodoc.FunctionDocumenter):
    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        # ensure that pipesegments are documented like functions
        if hasattr(member, "__pipesegment__"):
            return True
        return super().can_document_member(member, membername, isattr, parent)


class MethodDocumenter(BindingDocsMixin, autodoc.MethodDocumenter):
    pass


class PropertyDocumenter(autodoc.PropertyDocumenter):
    def get_doc(self):
        if self._new_docstrings is not None:
            return self._new_docstrings

        docstrings = super().get_doc()
        if not docstrings or "->" not in str(docstrings):
            func = self._get_property_getter()
            if func is None or getattr(func, "__doc__", None) is None:
                return docstrings
            gdoc = func.__doc__
            if any([gdoc.strip() in "\n".join(d).strip() for d in docstrings]):
                return docstrings
            docstrings.append(gdoc.split("\n"))

        for doclines in docstrings:
            for j, line in enumerate(doclines):
                if "->" in line:
                    self._cxx_type_ann = line.partition(" -> ")[-1].strip()
                    doclines[j] = ""
                    return docstrings

        return docstrings

    def add_directive_header(self, sig):
        # update property type annotation
        super().add_directive_header(sig)
        sourcename = self.get_sourcename()
        if hasattr(self, "_cxx_type_ann"):
            self.add_line("   :type: " + self._cxx_type_ann, sourcename)


# better formatting for enum name/value dicts
class AttributeDocumenter(autodoc.AttributeDocumenter):
    def import_object(self, raiseerror=False):
        self._is_enum_dict = False
        ret = super().import_object(raiseerror)
        if isinstance(self.object, dict):
            if self.objpath[-1] in ["names", "values"]:
                self._is_enum_dict = True
        return ret

    def should_suppress_value_header(self):
        if self._is_enum_dict:
            return True
        return super().should_suppress_value_header()

    def add_directive_header(self, sig):
        super().add_directive_header(sig)
        if self._is_enum_dict:
            sourcename = self.get_sourcename()
            self.add_line("   :type: dict", sourcename)

    def get_doc(self):
        if not self._is_enum_dict:
            return super().get_doc()

        if self.objpath[-1] == "names":
            txt = "Mapping of enum names to objects."
        else:
            txt = "Mapping of enum values to objects."

        v = pprint.pformat(self.object)
        v = textwrap.indent(v, "    ")
        return [[txt, "", ".. code-block::", "", ""] + v.split("\n")]


class ClassDocumenter(BindingDocsMixin, autodoc.ClassDocumenter):
    def add_line(self, line, source, *lineno):
        # remove empty bases directive
        if line.strip() == "Bases:":
            return
        super().add_line(line, source, *lineno)


def handle_bases(app, name, obj, options, bases: list):
    """
    Exclude generic base classes from inheritance lists
    """
    for b in list(bases):
        if 'pybind11' in str(b):
            bases.remove(b)
    if object in bases:
        bases.remove(object)


def setup(app):
    app.connect('autodoc-process-bases', handle_bases)
    app.add_autodocumenter(FunctionDocumenter, override=True)
    app.add_autodocumenter(MethodDocumenter, override=True)
    app.add_autodocumenter(PropertyDocumenter, override=True)
    app.add_autodocumenter(AttributeDocumenter, override=True)
    app.add_autodocumenter(ClassDocumenter, override=True)



#!/usr/bin/env python

import numpy as np
import healpy as hp
import os
from spt3g import core, maps
from astropy.wcs import WCS
from astropy.io import fits

for dim in [300, 301]:

    arr = np.arange(dim * dim).reshape(dim, dim)
    res = core.G3Units.arcmin
    deg = core.G3Units.deg
    a0 = 20 * deg
    d0 = -50 * deg
    d1 = -51 * deg
    x0 = 120
    y0 = 180

    verbose = False
    error = ''

    fm0 = maps.FlatSkyMap(dim, dim, res, proj=maps.MapProjection.ProjZEA,
                          alpha_center=a0, delta_center=d0,
                          x_center=x0, y_center=y0)
    fm0b = maps.FlatSkyMap(dim, dim, res, proj=maps.MapProjection.ProjZEA,
                           alpha_center=a0, delta_center=d1,
                           x_center=x0, y_center=y0)

    print('Checking projections for dim {}'.format(dim))
    if verbose:
        print('-' * 80)
    for p in [0, 1, 2, 3, 4, 5, 6, 7, 9]:
        if verbose:
            print('Checking Proj{}'.format(p))

        proj = getattr(maps.MapProjection, 'Proj{}'.format(p))
        fm1 = maps.FlatSkyMap(arr, res, proj=proj,
                              alpha_center=a0, delta_center=d0,
                              x_center=x0, y_center=y0)
        assert(np.allclose(fm1.angle_to_xy(a0, d0), (x0, y0)))
        assert(np.allclose(fm1.xy_to_angle(x0, y0), (a0, d0)))
        assert(fm1.alpha_center == a0)
        assert(fm1.delta_center == d0)
        assert(fm1.x_center == x0)
        assert(fm1.y_center == y0)

        fmc = fm0.clone(False)
        fmc.proj = proj
        assert(fmc.compatible(fm1))

        fmcb = fm0b.clone(False)
        fmcb.proj = proj
        fmc.delta_center = d1
        assert(fmcb.compatible(fmc))
        assert(np.all(np.array(fmcb.angle_to_xy(a0, d0)) == np.array(fmc.angle_to_xy(a0, d0))))

        maps.fitsio.save_skymap_fits('test_map.fits', fm1, overwrite=True)
        fm2 = maps.fitsio.load_skymap_fits('test_map.fits')['T']

        try:
            assert(fm1.compatible(fm2))
        except AssertionError:
            for attr in ['alpha_center', 'delta_center', 'x_center', 'y_center']:
                print(attr, getattr(fm1, attr), getattr(fm2, attr))
            error += '\nProj{}: fits error'.format(p)
        assert(np.allclose(fm1, fm2))
        assert(np.allclose(fm1.wcs.wcs.crpix, fm2.wcs.wcs.crpix))
        assert(np.allclose(fm1.wcs.wcs.crval, fm2.wcs.wcs.crval))
        assert(np.allclose(fm1.wcs.wcs.cdelt, fm2.wcs.wcs.cdelt))
        assert(np.allclose(fm1.wcs.wcs.pc, fm2.wcs.wcs.pc))

        fmask1 = maps.G3SkyMapMask(fm1, use_data=True)
        maps.fitsio.save_skymapmask_fits("test_mask.fits", fmask1, overwrite=True)
        fmask2 = maps.fitsio.load_skymapmask_fits("test_mask.fits")

        assert(fmask1.compatible(fmask2))
        assert((fmask1 == fmask2).all())

        w = fm2.wcs
        hdr = maps.fitsio.create_wcs_header(fm2)
        if verbose:
            print(repr(hdr))
        pixs = np.array([[0, 0], [0, dim - 1], [dim // 2, dim // 2], [dim - 1, 0], [dim - 1, dim - 1]]).astype(float)
        ra, dec = maps.get_ra_dec_map(fm2)

        angs = []
        bad = False
        for pix in pixs:
            wcs_ang = np.asarray(w.all_pix2world(pix[0], pix[1], 0))
            wcs_ang[wcs_ang > 180] -= 360
            g3_ang = np.asarray(fm2.xy_to_angle(*pix)) / deg
            idx = (int(pix[1]), int(pix[0]))
            try:
                try:
                    assert(np.allclose(g3_ang, [ra[idx] / deg, dec[idx] / deg]))
                except IndexError:
                    pass
                assert(np.allclose(wcs_ang, g3_ang))
                if verbose:
                    raise AssertionError
            except AssertionError:
                if not bad and not verbose:
                    print('ERROR: Proj{}'.format(p))
                    print('-' * 80)
                    print(repr(hdr))
                    print('-' * 80)
                bad = True
                print('pix', pix, 'pix2ang', [ra[idx] / deg, dec[idx] / deg],
                      'g3', g3_ang, 'wcs', wcs_ang, 'diff', np.abs(g3_ang - wcs_ang))
                error += '\nProj{}: xy_to_angle error'.format(p)
            angs.append(g3_ang * deg)

        for ang, pix in zip(angs, pixs):
            wcs_pix = np.asarray(w.all_world2pix(ang[0] / deg, ang[1] / deg, 0))
            g3_pix = np.asarray(fm2.angle_to_xy(*ang))
            try:
                assert(np.allclose(g3_pix, pix)) # round trip
                assert(np.allclose(wcs_pix, g3_pix))
                if verbose:
                    raise AssertionError
            except AssertionError:
                if not bad and not verbose:
                    print('ERROR: Proj{}'.format(p))
                    print('-' * 80)
                    print(repr(hdr))
                    print('-' * 80)
                bad = True
                print('ang', ang / deg, 'pix', pix, 'g3', g3_pix, 'wcs', wcs_pix,
                      'diff', np.abs(g3_pix - wcs_pix))
                error += '\nProj{}: angle_to_xy error'.format(p)

        if bad or verbose:
            print('-' * 80)

if error:
    os.remove('test_map.fits')
    os.remove('test_mask.fits')
    raise RuntimeError('Projection errors:{}'.format(error))

try:
    print('Checking Healpix')
    hm1 = maps.HealpixSkyMap(np.arange(12 * 64 * 64), pol_conv=maps.MapPolConv.COSMO)
    maps.fitsio.save_skymap_fits('test_map.fits', hm1, hm1, hm1, overwrite=True)
    hm2 = maps.fitsio.load_skymap_fits('test_map.fits')

    assert(hm2['T'].pol_conv == hm2['U'].pol_conv)
    assert(hm1.pol_conv == hm2['U'].pol_conv)
    assert(hm1.compatible(hm2['T']))
    assert(np.allclose(hm1, hm2['T']))
    assert(np.allclose(hm1, hm2['U']))

    print('Checking healpy.read_map')
    hm3 = hp.read_map('test_map.fits', dtype=float)
    hm3 = maps.HealpixSkyMap(hm3)
    assert(hm1.compatible(hm3))
    assert(np.allclose(hm1, hm3))

    print('Checking healpy.write_map')
    os.remove('test_map.fits')
    hp.write_map('test_map.fits', np.asarray(hm3), dtype=float, coord="C")
    hm4 = maps.fitsio.load_skymap_fits('test_map.fits')['T']
    assert(hm1.compatible(hm4))
    assert(np.allclose(hm1, hm4))

    hmask1 = maps.G3SkyMapMask(hm1, use_data=True)
    maps.fitsio.save_skymapmask_fits("test_mask.fits", hmask1, overwrite=True)
    hmask2 = maps.fitsio.load_skymapmask_fits("test_mask.fits")

    assert(hmask1.compatible(hmask2))
    assert((hmask1 == hmask2).all())

finally:
    if os.path.exists('test_map.fits'):
        os.remove('test_map.fits')
    if os.path.exists('test_mask.fits'):
        os.remove('test_mask.fits')
